"""Handling the mapping between ResSimpy attributes and the unit type of the attribute."""
from abc import ABC
from typing import Mapping

from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.Units.Units import UnitDimension


class BaseUnitMapping(ABC):
    """Base class for handling the mapping between ResSimpy attributes and the unit type of the attribute."""
    attribute_map: Mapping[str, UnitDimension]

    def __init__(self, unit_system: None | UnitSystem) -> None:
        self.unit_system = unit_system

    def get_unit_for_attribute(self, attribute_name: str, uppercase: bool = False) -> str:
        """Returns the unit variable for the given unit system.

        Args:
            attribute_name (str): name of the attribute to get the unit for
            uppercase (bool): if True returns the unit in uppercase
        """
        unit_dimension = self.attribute_map.get(attribute_name, None)
        if unit_dimension is None:
            raise AttributeError(f'Attribute {attribute_name} not recognised and does not have a unit definition')
        if self.unit_system is None:
            raise AttributeError('Unit system not defined')
        unit = unit_dimension.unit_system_enum_to_variable(unit_system=self.unit_system)
        if uppercase:
            unit = unit.upper()
        return unit

    def get_unit_for_keyword(self, keyword: str,
                             keyword_mapping: dict[str, tuple[str, type]],
                             uppercase: bool = False) -> str:
        """Given a simulator keyword, return the appropriate unit string.

        Args:
            keyword (str): Simulator keyword
            keyword_mapping (dict[str, tuple[str, type]]): Mapping of simulator keywords to attribute definitions
            uppercase (bool): if True returns the unit in uppercase

        Returns:
            str: unit string
        """
        attribute_name = keyword_mapping[keyword][0]  # Extract ResSimpy attribute name from provided keyword mapping
        return self.get_unit_for_attribute(attribute_name, uppercase)




EQUIL_KEYWORDS_VALUE_FLOAT = ['PINIT', 'DINIT', 'TINIT', 'GOC', 'WOC', 'PCGOC', 'PCWOC', 'PSAT', 'API', 'GWC',
                              'PCGWC', 'SALINITY', 'CHLORIDE', 'CALCIUM', 'LI_FACTOR', 'LI_AUTO',
                              'SORWMN', 'SORGMN', 'SGCMN', 'VAITS_TOLSG', 'VAITS_TOLSW',
                              'WOC_PALEO', 'GOC_PALEO']
EQUIL_KEYWORDS_VALUE_STR = ['OVERREAD', 'AUTOGOC_COMP', 'VIP_INIT']
EQUIL_OVERREAD_VALUES = ['PRESSURE', 'SW', 'SG', 'MOLEFRACTION']
EQUIL_SINGLE_KEYWORDS = ['WATERZONE_NEW_INIT', 'LI_AUTO_GAS', 'KEEPSG', 'KEEPSW', 'HONOR_GZONE', 'HONOR_WZONE',
                         'PCADJ_SCALING', 'NONEQ', 'SEDIMENTATION', 'MATCHVIPPRES', 'NOMATCHVIPPRES', 'CRINIT',
                         'HONOR_GASPRESSURE_GWC', 'POROSITY_INDEPENDENCE', 'POROSITY_DEPENDENCE']
EQUIL_INTSAT_KEYWORDS = ['INTSAT', 'VAITS']
EQUIL_TABLE_KEYWORDS = ['DEPTHVAR', 'OILMF', 'GASMF', 'COMPOSITION']
EQUIL_COMPOSITION_OPTIONS = ['X', 'Y']
EQUIL_TABLE_HEADERS = ['DEPTH', 'PSAT', 'TEMP', 'API', 'SALINITY', 'CHLORIDE', 'CALCIUM', 'X', 'Y']
EQUIL_KEYWORDS = EQUIL_KEYWORDS_VALUE_FLOAT + EQUIL_KEYWORDS_VALUE_STR + EQUIL_SINGLE_KEYWORDS + \
    EQUIL_INTSAT_KEYWORDS + EQUIL_TABLE_KEYWORDS



"""Handle Nexus files and preserve origin of include files."""
from __future__ import annotations

import os.path
from dataclasses import dataclass, field
from typing import Optional, Generator

# Use correct Self type depending upon Python version
import sys

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

from uuid import UUID
import re
import ResSimpy.Nexus.nexus_file_operations as nfo
import warnings
from ResSimpy.Nexus.NexusKeywords.structured_grid_keywords import GRID_OPERATION_KEYWORDS, GRID_ARRAY_FORMAT_KEYWORDS, \
    GRID_ARRAY_KEYWORDS
from ResSimpy.Utils.factory_methods import get_empty_list_str, get_empty_list_nexus_file, \
    get_empty_dict_uuid_list_int
from ResSimpy.File import File
import pathlib
import os
from datetime import datetime, timezone


@dataclass(kw_only=True, repr=True)
class NexusFile(File):
    """Class to deal with origin and structure of Nexus files and preserve origin of include files.

    Attributes:
        location (Optional[str]): Path to the original file being opened. Defaults to None.
        include_locations (Optional[list[str]]): list of file paths that the file contains. Defaults to None.
        origin (Optional[str]): Where the file was opened from. Defaults to None.
        include_objects (Optional[list[NexusFile]]): The include files but generated as a NexusFile instance. \
            Defaults to None.
        linked_user (Optional[str]): user or owner of the file. Defaults to None
        last_modified (Optional[datetime]): last modified date of the file
    """

    include_locations: Optional[list[str]] = field(default=None)
    origin: Optional[str] = None
    include_objects: Optional[list[NexusFile]] = field(default=None, repr=False)
    object_locations: Optional[dict[UUID, list[int]]] = field(default=None, repr=False)
    line_locations: Optional[list[tuple[int, UUID]]] = field(default=None, repr=False)
    linked_user: Optional[str] = field(default=None)
    last_modified: Optional[datetime] = field(default=None)

    def __init__(self, location: str,
                 include_locations: Optional[list[str]] = None,
                 origin: Optional[str] = None,
                 include_objects: Optional[list[NexusFile]] = None,
                 file_content_as_list: Optional[list[str]] = None,
                 linked_user: Optional[str] = None,
                 last_modified: Optional[datetime] = None) -> None:
        super().__init__(location=location, file_content_as_list=file_content_as_list)
        if origin is not None and location is not None:
            self.location = nfo.get_full_file_path(location, origin)
        else:
            self.location = location
        self.input_file_location: Optional[str] = location
        self.include_locations: Optional[list[str]] = get_empty_list_str() if include_locations is None else \
            include_locations
        self.origin: Optional[str] = origin
        self.include_objects: Optional[list[NexusFile]] = get_empty_list_nexus_file() \
            if include_objects is None else include_objects
        if self.object_locations is None:
            self.object_locations: dict[UUID, list[int]] = get_empty_dict_uuid_list_int()
        if self.line_locations is None:
            self.line_locations = []
        self.linked_user = linked_user
        self.last_modified = last_modified

    @classmethod
    def generate_file_include_structure(cls, file_path: str, origin: Optional[str] = None, recursive: bool = True,
                                        skip_arrays: bool = True, top_level_file: bool = True) -> Self:
        """Generates a nexus file instance for a provided text file with information storing the included files.

        Args:
            file_path (str): path to a file
            origin (Optional[str], optional): Where the file was opened from. Defaults to None.
            recursive (bool): Whether the method should recursively drill down multiple layers of include_locations.
            skip_arrays (bool): If set True skips the INCLUDE arrays that come after property array and VALUE
            top_level_file (bool): If set to True, the code assumes this is a 'top level' file rather than an included
            one.

        Returns:
            NexusFile: a class instance for NexusFile with knowledge of include files
        """
        def __get_pathlib_path_details(full_file_path: str):
            if full_file_path == "" or full_file_path is None:
                return None
            pathlib_path = pathlib.Path(full_file_path)
            owner: str = ''
            group: str = ''
            try:
                owner = pathlib_path.owner()  # type: ignore
                group = pathlib_path.group()  # type: ignore
            except NotImplementedError:
                # owner or group not supported on this system, continue without filling out that information
                pass
            except PermissionError:
                # user doesn't have permission to access the file, continue without filling out that information
                warnings.warn(f'PermissionError when trying to access file at {full_file_path}')
            except FileNotFoundError:
                # file not found, continue without filling out that information
                warnings.warn(f'FileNotFoundError when trying to access file at {full_file_path}')
            except KeyError:
                # Group or owner doesn't exist on this system, continue without filling out that information
                warnings.warn(f'Unable to find the group for the file at {full_file_path}')

            if owner is not None and group is not None:
                return f"{owner}:{group}"
            elif owner is not None:
                return owner
            return None

        def __get_datetime_from_os_stat(full_file_path: str):
            if full_file_path == "" or full_file_path is None:
                return None
            stat_obj = os.stat(full_file_path)
            timestamp = stat_obj.st_mtime
            if timestamp is None:
                return None
            return datetime.fromtimestamp(timestamp, tz=timezone.utc)

        full_file_path = file_path
        if origin is not None:
            full_file_path = nfo.get_full_file_path(file_path, origin)

        try:
            file_as_list = nfo.load_file_as_list(full_file_path)
        except FileNotFoundError:
            # handle if a file can't be found
            location = file_path

            nexus_file_class = cls(location=location,
                                   include_locations=None,
                                   origin=origin,
                                   include_objects=None,
                                   file_content_as_list=None,
                                   linked_user=None,
                                   last_modified=None)
            warnings.warn(UserWarning(f'No file found for: {file_path} while loading {origin}'))
            return nexus_file_class

        # check last modified and user for the file
        user = __get_pathlib_path_details(full_file_path)
        last_changed = __get_datetime_from_os_stat(full_file_path)

        # prevent python from mutating the lists that it's iterating over
        modified_file_as_list: list[str] = []
        # search for the INCLUDE keyword and append to a list:
        inc_file_list: list[str] = []
        includes_objects: Optional[list[NexusFile]] = []
        skip_next_include = False
        previous_line: str

        for i, line in enumerate(file_as_list):
            if len(modified_file_as_list) >= 1:
                previous_line = modified_file_as_list[len(modified_file_as_list) - 1].rstrip('\n')
                if previous_line.endswith('>'):
                    modified_file_as_list[len(modified_file_as_list) - 1] = previous_line[:-1] + line
                else:
                    modified_file_as_list.append(line)
            else:
                modified_file_as_list.append(line)
            if line.rstrip('\n').endswith('>'):
                continue
            if nfo.check_token("INCLUDE", line):
                # Include found, check if we should skip loading it in (e.g. if it is a large array file)
                ignore_keywords = ['NOLIST']
                previous_value = nfo.get_previous_value(file_as_list=file_as_list[0: i + 1], search_before='INCLUDE',
                                                        ignore_values=ignore_keywords)

                keywords_to_skip_include = GRID_ARRAY_FORMAT_KEYWORDS + GRID_OPERATION_KEYWORDS + ["CORP"]
                if previous_value is None:
                    skip_next_include = False

                elif previous_value.upper() in keywords_to_skip_include:
                    skip_next_include = True

            elif nfo.check_token("VALUE", line) and not top_level_file:
                # Check if this is an 'embedded' grid array file. If it is, return this file with only the content up
                # to this point to help with performance when analysing the files.
                previous_value = nfo.get_previous_value(file_as_list=file_as_list[0: i + 1], search_before='VALUE')
                next_value = nfo.get_next_value(start_line_index=0, file_as_list=file_as_list[i:],
                                                search_string=line.upper().split('VALUE')[1])

                if previous_value is None or next_value is None:
                    continue

                if next_value.upper() != 'INCLUDE' and previous_value.upper() in GRID_ARRAY_KEYWORDS:
                    nexus_file_class = cls(
                        location=file_path,
                        include_locations=inc_file_list,
                        origin=origin,
                        include_objects=includes_objects,
                        file_content_as_list=modified_file_as_list
                    )

                    return nexus_file_class
                else:
                    continue

            else:
                continue
            inc_file_path = nfo.get_token_value('INCLUDE', line, file_as_list)
            if inc_file_path is None:
                continue
            inc_full_path = nfo.get_full_file_path(inc_file_path, origin=full_file_path)
            # store the included files as files inside the object
            inc_file_list.append(inc_full_path)
            if not recursive:
                continue
            elif skip_arrays and skip_next_include:
                inc_file = cls(location=inc_file_path,
                               include_locations=None,
                               origin=full_file_path,
                               include_objects=None,
                               file_content_as_list=None,
                               linked_user=user,
                               last_modified=last_changed)
                if includes_objects is None:
                    raise ValueError('include_objects is None - recursion failure.')
                skip_next_include = False
            else:
                inc_file = cls.generate_file_include_structure(inc_file_path, origin=full_file_path, recursive=True,
                                                               skip_arrays=skip_arrays, top_level_file=False)
                if includes_objects is None:
                    raise ValueError('include_objects is None - recursion failure.')

            includes_objects.append(inc_file)

        includes_objects = None if not includes_objects else includes_objects

        nexus_file_class = cls(
            location=file_path,
            include_locations=inc_file_list,
            origin=origin,
            include_objects=includes_objects,
            file_content_as_list=modified_file_as_list,
            linked_user=user,
            last_modified=last_changed
        )

        return nexus_file_class

    def export_network_lists(self):
        """Exports lists of connections from and to for use in network graphs.

        Raises:
            ValueError: If the from and to lists are not the same length

        Returns:
            tuple[list]: list of to and from file paths where the equivalent indexes relate to a connection
        """
        from_list = [self.origin]
        to_list = [self.location]
        if not [self.origin]:
            to_list = []
        if self.include_locations is not None:
            from_list += [self.location] * len(self.include_locations)
            to_list += self.include_locations
        if len(from_list) != len(to_list):
            raise ValueError(
                f"{from_list=} and {to_list=} are not the same length")

        return from_list, to_list

    @dataclass
    class FileIndex:
        index: int

    def iterate_line(self, file_index: Optional[FileIndex] = None, max_depth: Optional[int] = None,
                     parent: Optional[NexusFile] = None, prefix_line: Optional[str] = None,
                     keep_include_references=False) -> \
            Generator[str, None, None]:
        """Generator object for iterating over a list of strings with nested NexusFile objects in them.

        Yields:
            str: sequential line from the file.
        """

        if file_index is None:
            file_index = NexusFile.FileIndex(index=0)
        if parent is None:
            parent = self
            parent.line_locations = []
        if parent.line_locations is None:
            parent.line_locations = []
        if prefix_line is not None and prefix_line != ' ':
            file_index.index += 1
            yield prefix_line

        new_entry = (file_index.index, self.id)
        if new_entry not in parent.line_locations:
            parent.line_locations.append(new_entry)
        depth: int = 0
        if max_depth is not None:
            depth = max_depth
        if self.file_content_as_list is None:
            warnings.warn(f'No file content found for file: {self.location}')
            return
        for row in self.file_content_as_list:
            if nfo.check_token('INCLUDE', row):
                incfile_location = nfo.get_token_value('INCLUDE', row, self.file_content_as_list)
                if incfile_location is None:
                    continue
                split_line = re.split(incfile_location, row, maxsplit=1, flags=re.IGNORECASE)
                if len(split_line) == 2:
                    prefix_line, suffix_line = split_line
                    prefix_line = re.sub('INCLUDE', '', prefix_line, flags=re.IGNORECASE)
                    prefix_line = prefix_line.rstrip() + ' '
                    suffix_line = suffix_line.lstrip()
                else:
                    prefix_line = row.replace(incfile_location, '')
                    suffix_line = None

                include_file = None
                if self.include_objects is None:
                    raise ValueError(f'No include objects found in the nexusfile to expand over for file: '
                                     f'{self.location}')
                for obj in self.include_objects:
                    if obj.location == incfile_location:
                        include_file = obj
                        break
                    if self.origin is not None and \
                            obj.location == nfo.get_full_file_path(incfile_location, self.origin):
                        include_file = obj
                        break
                    if obj.location is not None and \
                            os.path.basename(obj.location) == os.path.basename(incfile_location):
                        include_file = obj
                        break

                if (max_depth is None or depth > 0) and include_file is not None:
                    level_down_max_depth = None if max_depth is None else depth - 1

                    if keep_include_references:
                        yield row

                    yield from include_file.iterate_line(file_index=file_index, max_depth=level_down_max_depth,
                                                         parent=parent, prefix_line=prefix_line)

                    new_entry = (file_index.index, self.id)
                    if new_entry not in parent.line_locations:
                        parent.line_locations.append(new_entry)
                    if suffix_line:
                        file_index.index += 1
                        # Add in space between include location and the rest of the line
                        suffix_line = ' ' + suffix_line
                        yield suffix_line
                else:
                    continue
            else:
                file_index.index += 1
                yield row

    @property
    def get_flat_list_str_file(self) -> list[str]:
        if self.file_content_as_list is None:
            raise ValueError(f'No file content found for {self.location}')
        flat_list = list(self.iterate_line(file_index=None, keep_include_references=False))
        return flat_list

    @property
    def get_flat_list_str_file_including_includes(self) -> list[str]:
        if self.file_content_as_list is None:
            raise ValueError(f'No file content found for {self.location}')
        flat_list = list(self.iterate_line(file_index=None, keep_include_references=True))
        return flat_list

    # TODO write an output function using the iterate_line method
    def get_full_network(self, max_depth: Optional[int] = None) -> tuple[list[str | None], list[str]]:
        """Recursively constructs two lists of from and to nodes representing the connections between files.

        Args:
            max_depth (Optional[int], optional): depth of the iteration to construct the network down to. \
                Defaults to None.

        Returns:
            tuple[list[str | None], list[str | None]]: two lists of from and to nodes where corresponding \
                indices create an edge within a graph network. e.g. (from_list[i], to_list[i]) \
                is a connection between two files.
        """
        depth: int = 0
        from_list = [self.origin]
        to_list = [self.location]
        if max_depth is not None:
            depth = max_depth
        if self.file_content_as_list is None:
            return from_list, to_list
        for row in self.file_content_as_list:
            if isinstance(row, NexusFile):
                if (max_depth is None or depth > 0):
                    level_down_max_depth = None if max_depth is None else depth - 1
                    temp_from_list, temp_to_list = row.export_network_lists()
                    from_list.extend(temp_from_list)
                    to_list.extend(temp_to_list)
                    temp_from_list, temp_to_list = row.get_full_network(max_depth=level_down_max_depth)
                    from_list.extend(temp_from_list)
                    to_list.extend(temp_to_list)
        return from_list, to_list

    def add_object_locations(self, obj_uuid: UUID, line_indices: list[int]) -> None:
        """Adds a uuid to the object_locations dictionary. Used for storing the line numbers where objects are stored
        within the flattened file_as_list.

        Args:
            obj_uuid (UUID): unique identifier of the object being created/stored.
            line_indices (list[int]): line number in the flattened file_content_as_list
                (i.e. from the get_flat_list_str_file method).
        """
        if self.object_locations is None:
            self.object_locations: dict[UUID, list[int]] = get_empty_dict_uuid_list_int()
        existing_line_locations = self.object_locations.get(obj_uuid, None)
        if existing_line_locations is not None:
            existing_line_locations.extend(line_indices)
            existing_line_locations.sort()
        else:
            self.object_locations[obj_uuid] = line_indices

    def __update_object_locations(self, line_number: int, number_additional_lines: int) -> None:
        """Updates the object locations in a nexusfile by the additional lines. Used when files have been modified and
        an addition/removal of lines has occurred. Ensures that the object locations are correct to the actual lines
        in the file_as_list.

        Args:
            line_number (int): Line number at which the new lines have been added
            number_additional_lines (int): number of new lines added.
        """
        if self.object_locations is None:
            return
        for object_id, list_indices in self.object_locations.items():
            for i, obj_index in enumerate(list_indices):
                if obj_index >= line_number:
                    self.object_locations[object_id][i] = obj_index + number_additional_lines

    def __remove_object_locations(self, obj_uuid: UUID) -> None:
        """Removes an object location based on the obj_uuid provided. Used when removing objects in the file_as_list.

        Args:
            obj_uuid (UUID): id of the removed object.
        """
        if self.object_locations is None:
            raise ValueError(f'No object locations found for file {self.location}')

        if self.object_locations.get(obj_uuid, None) is None:
            raise ValueError(f'No object with {obj_uuid=} found within the object locations')
        self.object_locations.pop(obj_uuid, None)

    def find_which_include_file(self, flattened_index: int) -> tuple[File, int]:
        """Given a line index that relates to a position within the flattened file_as_list from the method
        get_flat_file_as_list.

        Args:
            flattened_index (int): index in the flattened file as list structure

        Returns:
            tuple[File, int] where the first element is the file that the relevant line is in and the second
            element is the relative index in that file.
        """
        if self.line_locations is None:
            # call get_flat_list_str_file to ensure line locations are updated
            self.get_flat_list_str_file
            if self.line_locations is None:
                raise ValueError("No include line locations found.")

        line_locations = [x[0] for x in self.line_locations]
        line_locations.sort()
        uuid_index: Optional[UUID] = None
        index_in_included_file = 0

        # Find the Nexusfile containing the line we are looking for
        for numlines, obj_id in self.line_locations:
            if numlines <= flattened_index:
                uuid_index = obj_id
            if numlines >= flattened_index:
                break

        get_next_line_count = False
        lines_already_included = 0
        previous_lines_value = 0

        # Add the previously included lines from the file (if any) to the location after it has been split
        for numlines, obj_id in self.line_locations:
            if obj_id == uuid_index:
                get_next_line_count = True
                previous_lines_value = numlines
                continue
            else:
                # If we have gone beyond where the line is, calculate the location in the relevant file
                if numlines >= flattened_index:
                    lines_already_included += flattened_index - previous_lines_value
                    break

                if get_next_line_count:
                    lines_already_included += numlines - previous_lines_value
                    get_next_line_count = False
                previous_lines_value = numlines

        if flattened_index > line_locations[-1]:
            lines_already_included += flattened_index - line_locations[-1]

        index_in_included_file += lines_already_included

        if uuid_index == self.id or self.include_objects is None:
            return self, index_in_included_file

        nexus_file = None
        for file in self.include_objects:
            if file.id == uuid_index:
                nexus_file = file
            elif file.include_objects is not None:
                # CURRENTLY THIS ONLY SUPPORTS 2 LEVELS OF INCLUDES
                for lvl_2_include in file.include_objects:
                    if lvl_2_include.id == uuid_index:
                        nexus_file = lvl_2_include
        if nexus_file is None:
            raise ValueError(f'No file with {uuid_index=} found within include objects')

        return nexus_file, index_in_included_file

    def add_to_file_as_list(self, additional_content: list[str], index: int,
                            additional_objects: Optional[dict[UUID, list[int]]] = None,
                            comments: Optional[str] = None) -> None:
        """To add content to the file as list, also updates object numbers and optionally allows user \
        to add several additional new objects.

        Args:
            additional_content (list[str]): Additional lines as a list of strings to be added.
            index (int): index to insert the new lines at in the calling flat_file_as_list
            additional_objects (Optional[dict[UUID, int]]): defaults to None. Otherwise, a dictionary keyed with the \
            UUID of the new objects to add as well as the corresponding index of the object in the original \
            calling NexusFile
            comments (str | None): defaults to None. Comments to add in-line to the file.
        """
        if comments is not None:
            additional_content = NexusFile.insert_comments(additional_content, comments)

        nexusfile_to_write_to, relative_index = self.find_which_include_file(index)
        if nexusfile_to_write_to.file_content_as_list is None:
            raise ValueError(f'No file content to write to in file: {nexusfile_to_write_to}')
        nexusfile_to_write_to.file_content_as_list = \
            nexusfile_to_write_to.file_content_as_list[:relative_index] + \
            additional_content + nexusfile_to_write_to.file_content_as_list[relative_index:]

        self._file_modified_set(True)

        # update object locations
        self.__update_object_locations(line_number=index, number_additional_lines=len(additional_content))

        if additional_objects is None:
            return
        for object_id, line_index in additional_objects.items():
            self.add_object_locations(obj_uuid=object_id, line_indices=line_index)

    def remove_object_from_file_as_list(self, objects_to_remove: list[UUID]) -> None:
        """Removes all associated lines in the file as well as the object locations relating to a list of objects."""
        if self.object_locations is None:
            raise ValueError('Cannot remove object from object_locations as object_locations is None. '
                             'Check object locations is being populated properly.')
        for obj_to_remove in objects_to_remove:
            # find all locations in the code that relate to the object
            obj_locs = self.object_locations.get(obj_to_remove, None)
            if obj_locs is None:
                continue
            # sort from highest to lowest to ensure line indices are not affected by removal of lines
            sorted_obj_locs = sorted(obj_locs, reverse=True)
            for i, index in enumerate(sorted_obj_locs):
                if i == 0:
                    # for the first removal remove the object location
                    self.remove_from_file_as_list(index, objects_to_remove=[obj_to_remove])
                else:
                    # the remaining iterations remove just the lines
                    self.remove_from_file_as_list(index)
        self._file_modified_set(True)

    def remove_from_file_as_list(self, index: int, objects_to_remove: Optional[list[UUID]] = None,
                                 string_to_remove: Optional[str] = None) -> None:
        """Remove an entry from the file as list. Also updates existing object locations and removes any \
        specified objects from the object locations dictionary.

        Args:
            index (int): index n the calling flat_file_as_list to remove the entry from
            objects_to_remove (Optional[list[UUID]]): list of object id's to remove from the object locations. \
            Defaults to None
            string_to_remove (Optional[str]): if specified will only remove the listed string from the entry \
            at the index. Defaults to None, which removes the entire entry.

        """
        nexusfile_to_write_to, relative_index = self.find_which_include_file(index)

        # remove the line in the file:
        if nexusfile_to_write_to.file_content_as_list is None:
            raise ValueError(
                f'No file content in the file attempting to remove line from {nexusfile_to_write_to.location}')

        if string_to_remove is None:
            nexusfile_to_write_to.file_content_as_list.pop(relative_index)
            self.__update_object_locations(line_number=index, number_additional_lines=-1)
        else:
            entry_to_replace = nexusfile_to_write_to.file_content_as_list[relative_index]
            if isinstance(entry_to_replace, str):
                nexusfile_to_write_to.file_content_as_list[relative_index] = \
                    entry_to_replace.replace(string_to_remove, '', 1)
            else:
                raise ValueError(
                    f'Tried to replace at non string value at index: {relative_index} in '
                    f'file_as_list instead got {entry_to_replace}')

        if objects_to_remove is not None:
            for object_id in objects_to_remove:
                self.__remove_object_locations(object_id)
        self._file_modified_set(True)

    @staticmethod
    def insert_comments(additional_content: list[str], comments: str) -> list[str]:
        """Adds comments alongside additional content.

        Args:
            additional_content (list[str]): additional lines of the file to be added with a new entry per line.
            comments (str): comments to be added to all lines

        Returns:
            list of strings within the content.
        """
        for index, element in enumerate(additional_content):
            newline_index = element.find('\n')
            if newline_index != -1:
                modified_text = element[:newline_index] + ' ! ' + comments + element[newline_index:]
                additional_content[index] = modified_text
                return additional_content

        additional_content[-1] += ' ! ' + comments + '\n'
        return additional_content

    def get_object_locations_for_id(self, id: UUID) -> list[int]:
        """Gets the number of object locations for a specified id."""
        if self.object_locations is None or len(self.object_locations[id]) == 0:
            raise ValueError(f'No object locations specified, cannot find id: {id} in {self.object_locations}')
        return self.object_locations[id]

    def write_to_file(self, new_file_path: None | str = None, write_includes: bool = False,
                      write_out_all_files: bool = False, overwrite_file: bool = False) -> None:
        """Writes to file specified in self.location the strings contained in the list self.file_content_as_list.

        Args:
            new_file_path (None | str): writes to self.location if left as None. Otherwise writes to new_file_name.
            write_includes (bool): If True will write out all include files within the file. Defaults to False.
            write_out_all_files (bool): If False will write only modified files. Otherwise will write all files.
            overwrite_file (bool): If True will overwrite the file at the location specified by new_file_path. \
            Otherwise will raise an error if the file already exists.
        """
        # overwrite File base class method to allow for write_includes
        if new_file_path is None and overwrite_file:
            # In this case just overwrite the file with the existing path:
            new_file_path = self.location
        elif new_file_path is None and not overwrite_file:
            raise ValueError('No file path to write to, and overwrite_file set to False')
        if self.file_content_as_list is None:
            raise ValueError(f'No file data to write out, instead found {self.file_content_as_list}')

        if new_file_path is None:
            raise ValueError('No file path to write to.')
        if os.path.exists(new_file_path) and not overwrite_file:
            raise ValueError(f'File already exists at {new_file_path} and overwrite_file set to False')

        # create directories that do not exist
        if not os.path.exists(os.path.dirname(new_file_path)):
            os.makedirs(os.path.dirname(new_file_path))

        if write_includes and self.include_objects is not None:
            for file in self.include_objects:
                write_file: bool = file.file_modified or write_out_all_files
                write_file = write_file or (new_file_path != self.location and not os.path.isabs(file.location))
                if file.file_content_as_list is None:
                    warnings.warn(f'No content found for file: {file}. Not writing file.')
                    continue
                new_root_name = f'{os.path.basename(new_file_path).split(".")[0]}_{os.path.basename(file.location)}'
                # write the include file to the same directory.
                include_file_name = os.path.join(os.path.dirname(new_file_path), new_root_name)
                if write_file:
                    self.update_include_location_in_file_as_list(include_file_name, file)
                if write_file:
                    file.write_to_file(include_file_name, write_includes=True, write_out_all_files=write_out_all_files,
                                       overwrite_file=overwrite_file)

        file_str = ''.join(self.file_content_as_list)

        # update the location:
        if self.file_modified or write_out_all_files:
            self.location = new_file_path
            with open(new_file_path, 'w') as fi:
                fi.write(file_str)
            # reset the modified file state
            self._file_modified_set(False)

    def update_include_location_in_file_as_list(self, new_path: str, include_file: NexusFile) -> None:
        """Updates the path of an include file within this file's file_as_list.

        Args:
            new_path (str): Updates the path in the file as list to point towards the new include location.
            include_file (NexusFile): include object whose path is being modified
        """
        # try and find the path of the file that should be replaced (i.e. how it is currently written in the file)
        if self.include_locations is None or not self.include_locations or include_file.input_file_location is None:
            raise ValueError('No include locations found and therefore cannot update include path')
        file_path_to_replace = include_file.input_file_location
        file_content = self.file_content_as_list
        if file_content is None or not file_content:
            raise ValueError(f'No file content found within file {self.location}')

        for line in file_content:
            if not nfo.check_token('INCLUDE', line):
                continue
            # if the right path to replace is found then replace it
            if nfo.get_expected_token_value('INCLUDE', line, file_content) == file_path_to_replace:
                nfo.get_expected_token_value('INCLUDE', line, file_content, replace_with=new_path)
                self._file_modified_set(True)
        # replace the location in the include locations list using the original path

        index_of_path_to_replace = self.include_locations.index(include_file.location)
        # get the full path and update it in the include file object
        # TODO maybe include this in a setter attr for the location.
        include_file.location = nfo.get_full_file_path(new_path, self.location)

        self.include_locations[index_of_path_to_replace] = include_file.location
        # update the new path
        include_file.input_file_location = new_path



# TODO: Not a complete list. Check manual for missing keywords.
# TODO: Delete the keywords that are not tokens (i.e. that are only values in a table)

VALVE_TABLE_KEYWORDS = ['VALVE', 'ICV', 'CHOKE']

VALVE_RATE_KEYWORDS = ['QALL', 'QOIL', 'QGAS', 'QWATER', 'QLIQ']

VALVE_KEYWORDS = VALVE_TABLE_KEYWORDS



from abc import ABC
from dataclasses import dataclass, field
from typing import Literal
from ResSimpy.OperationsMixin import NetworkOperationsMixIn
from ResSimpy.Wellbore import Wellbore


@dataclass(kw_only=True, repr=False)
class Wellbores(NetworkOperationsMixIn, ABC):
    __connections: list[Wellbore] = field(default_factory=list)

    @property
    def _network_element_name(self) -> Literal['wellbores']:
        return 'wellbores'



"""The Nexus Simulator class is the main class for interacting with Nexus models."""
from __future__ import annotations

import os
import warnings
from typing import Any, Union, Optional

import resqpy.model as rq
from datetime import datetime
import ResSimpy.Nexus.nexus_file_operations as nfo
from ResSimpy.Nexus.DataModels.FcsFile import FcsNexusFile
from ResSimpy.Nexus.DataModels.NexusFile import NexusFile
from ResSimpy.Nexus.NexusPVTMethods import NexusPVTMethods
from ResSimpy.Nexus.NexusSeparatorMethods import NexusSeparatorMethods
from ResSimpy.Nexus.NexusWaterMethods import NexusWaterMethods
from ResSimpy.Nexus.NexusEquilMethods import NexusEquilMethods
from ResSimpy.Nexus.NexusRockMethods import NexusRockMethods
from ResSimpy.Nexus.NexusRelPermMethods import NexusRelPermMethods
from ResSimpy.Nexus.NexusValveMethods import NexusValveMethods
from ResSimpy.Nexus.NexusAquiferMethods import NexusAquiferMethods
from ResSimpy.Nexus.NexusHydraulicsMethods import NexusHydraulicsMethods
from ResSimpy.Nexus.NexusGasliftMethods import NexusGasliftMethods
from ResSimpy.Nexus.DataModels.StructuredGrid.NexusGrid import NexusGrid
from ResSimpy.Nexus.NexusEnums.DateFormatEnum import DateFormat
from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.Nexus.NexusNetwork import NexusNetwork
from ResSimpy.Nexus.NexusReporting import Reporting
from ResSimpy.Nexus.NexusWells import NexusWells
from ResSimpy.Nexus.runcontrol_operations import SimControls
from ResSimpy.Nexus.logfile_operations import Logging
from ResSimpy.Nexus.structured_grid_operations import StructuredGridOperations
from ResSimpy.Simulator import Simulator


class NexusSimulator(Simulator):

    def __init__(self, origin: Optional[str] = None, destination: Optional[str] = None,
                 root_name: Optional[str] = None, nexus_data_name: str = "data", write_times: bool = False,
                 manual_fcs_tidy_call: bool = False, lazy_loading: bool = True) -> None:
        """Nexus simulator class. Inherits from the Simulator super class.

        Args:
            origin (Optional[str], optional): file path to the fcs file. Defaults to None.
            root_name (Optional[str], optional): Root file name of the fcs. Defaults to None.
            nexus_data_name (str, optional): Folder name for the nexus data files to be stored in. Defaults to "data".
            write_times (bool, optional): Sets whether the runcontrol file will expand the include files with time \
                cards in. Defaults to True.
            manual_fcs_tidy_call (bool, optional): Determines whether fcs_tidy should be called - Currently not used. \
                Defaults to False.

        Attributes:
            run_control_file_path (Optional[str]): file path to the run control file - derived from the fcs file
            __destination (Optional[str]): output path for the simulation. Currently not used.
            date_format (DateFormat): Enum value representing the date format.
            __original_fcs_file_path (str): Path to the original fcs file path supplied
            __new_fcs_file_path (str): Where the new fcs will be saved to
            __force_output (bool): private attribute of force_output
            __origin (str): private attribute of origin. File path to the fcs file.
            __root_name (str): private attribute of root_name. Root file name of the fcs.
            __nexus_data_name (str): private attribute of nexus_data_name. Folder name for the nexus data files to be \
                stored in.
            __structured_grid_file_path (Optional[str]): file path to the structured grid.
            __structured_grid_file (Optional[NexusGrid]): StructuredGridFile object representing the \
                structured grid used in Nexus
            __run_units (Optional[str]): Unit system used in the Nexus model
            use_american_run_units (bool): True if an American unit system is used equivalent to 'ENGLISH'. \
                False otherwise. For the RUN_UNITS keyword.
            use_american_input_units (bool): True if an American unit system is used equivalent to 'ENGLISH'. \
                False otherwise. For the DEFAULT_UNITS keyword.
            __write_times (bool): private attribute for write_times. Sets whether the runcontrol file will expand \
                the include files with time cards in.
            __manual_fcs_tidy_call (bool): private attribute for manual_fcs_tidy_call. Determines whether fcs_tidy \
                should be called

        Raises:
            ValueError: If the FCS file path is not given

        Examples:
            Example of calling a NexusSimulator with a file path to an fcs file:

            >>> from ResSimpy.Nexus.NexusSimulator import NexusSimulator
            >>> model = NexusSimulator(origin="/path/to/fcs_file.fcs")

        """
        if origin is None:
            raise ValueError(f'Origin path to model fcs file is required. Instead got {origin}.')
        self.origin: str = origin

        self._start_date: str = ''
        self.run_control_file_path: Optional[str] = ''
        self.__destination: Optional[str] = None
        self.date_format: DateFormat = DateFormat.MM_DD_YYYY  # Nexus default
        self.__original_fcs_file_path: str = self.origin
        self.__new_fcs_file_path: str = self.origin
        self.__nexus_data_name: str = nexus_data_name
        self.__run_units: UnitSystem = UnitSystem.ENGLISH  # The Nexus default
        self.root_name: str = root_name
        self.use_american_run_units: bool = False
        self.use_american_input_units: bool = False
        self.__write_times: bool = write_times
        self.__manual_fcs_tidy_call: bool = manual_fcs_tidy_call

        self.__default_units: UnitSystem = UnitSystem.ENGLISH  # The Nexus default
        #
        self._network: NexusNetwork = NexusNetwork(model=self)
        self._wells: NexusWells = NexusWells(self)
        self._grid: Optional[NexusGrid] = None
        # Model dynamic properties
        self._pvt: NexusPVTMethods = NexusPVTMethods(model_unit_system=self.default_units)
        self._separator: NexusSeparatorMethods = NexusSeparatorMethods(model_unit_system=self.default_units)
        self._water: NexusWaterMethods = NexusWaterMethods(model_unit_system=self.default_units)
        self._equil: NexusEquilMethods = NexusEquilMethods(model_unit_system=self.default_units)
        self._rock: NexusRockMethods = NexusRockMethods(model_unit_system=self.default_units)
        self._relperm: NexusRelPermMethods = NexusRelPermMethods(model_unit_system=self.default_units)
        self._valve: NexusValveMethods = NexusValveMethods(model_unit_system=self.default_units)
        self._aquifer: NexusAquiferMethods = NexusAquiferMethods(model_unit_system=self.default_units)
        self._hydraulics: NexusHydraulicsMethods = NexusHydraulicsMethods(model_unit_system=self.default_units)
        self._gaslift: NexusGasliftMethods = NexusGasliftMethods(model_unit_system=self.default_units)
        # Nexus operations modules
        self._sim_controls: SimControls = SimControls(self)
        self.reporting: Reporting = Reporting(self)
        self._structured_grid_operations: StructuredGridOperations = StructuredGridOperations(self)
        self.logging: Logging = Logging(self)
        self.__lazy_loading: bool = lazy_loading

        if destination is not None and destination != '':
            self.set_output_path(path=destination.strip())

        # Check the status of any existing or completed runs related to this model
        self.get_simulation_status(from_startup=True)

        self._model_files: FcsNexusFile
        # Load in the model
        self.__load_fcs_file()

    def remove_temp_from_properties(self):
        """Updates model values if the files are moved from a temp directory
        Replaces the first instance of temp/ in the file paths in the nexus simulation file paths.

        Raises:
            ValueError: if any of [__structured_grid_file_path, __new_fcs_file_path, __surface_file_path] are None.
        """
        if self.model_files.structured_grid_file.location is None:
            raise ValueError(
                "No structured_grid_file_path found, can't remove temporary properties from file path")
        if self.__new_fcs_file_path is None:
            raise ValueError(
                "No __new_fcs_file_path found, can't remove temporary properties from file path")
        if self.model_files.surface_files[1] is None or self.model_files.surface_files[1].location is None:
            raise ValueError(
                "No __surface_file_path found, can't remove temporary properties from file path")

        self._origin = self._origin.replace('temp/', '', 1)
        self.__root_name = self.__root_name.replace('temp/', '', 1)
        self.model_files.structured_grid_file.location = \
            self.model_files.structured_grid_file.location.replace('temp/', '', 1)
        self.__new_fcs_file_path = self.__new_fcs_file_path.replace('temp/', '', 1)
        self.model_files.surface_files[1].location = self.model_files.surface_files[1].location.replace('temp/', '', 1)

    def get_simulation_status(self, from_startup: bool = False) -> Optional[str]:
        return self.logging.get_simulation_status(from_startup)

    def get_simulation_progress(self) -> float:
        return self.logging.get_simulation_progress()

    def get_users_linked_with_files(self) -> Optional[list[tuple[Optional[str], Optional[str], Optional[datetime]]]]:
        return self.model_files.files_info

    @property
    def model_files(self) -> FcsNexusFile:
        return self._model_files

    @property
    def network(self) -> NexusNetwork:
        return self._network

    @property
    def structured_grid_path(self):
        """Returns the location of the structured grid file."""
        return self.model_files.structured_grid_file.location

    @property
    def default_units(self):
        """Returns the default units."""
        return self.__default_units

    @property
    def run_units(self):
        """Returns the run units."""
        return self.__run_units

    @property
    def new_fcs_name(self):
        """Returns the new name for the FCS file without the fcs extension."""
        return self.__root_name

    @property
    def write_times(self) -> bool:
        return self.__write_times

    @property
    def original_fcs_file_path(self):
        return self.__original_fcs_file_path

    @property
    def root_name(self):
        return self.__root_name

    @root_name.setter
    def root_name(self, value: str) -> None:
        """Returns the name of the fcs file without the .fcs extension.

        Returns:
            str: string of the fcs file without the .fcs extension.
        """
        if value is not None:
            rootname = value
        else:
            rootname = os.path.basename(self._origin)
            rootname = rootname.split(".fcs")[0]
        self.__root_name = rootname

    @staticmethod
    def get_check_run_input_units_for_models(models: list[str]) -> tuple[Optional[bool], Optional[bool]]:
        # TODO: add LAB units
        """Returns the run and input unit formats for all the supplied models.

        Supported model formats:
            RESQML type epc files ending in ".epc"
            Nexus files containing a line identifying the "RUN_UNITS" or "DEFAULT_UNITS".

        Supported units: ENGLISH, METRIC

        Args:
            models (list[str]): list of paths to supported reservoir models

        Raises:
            ValueError: if a model in the list is using inconcistent run/default units

        Returns:
            Tuple[Optional[Bool], Optional[Bool]]: If all units are consistent between models,
                Returns (True, True) if 'ft' is the length unit in an epc or Nexus specifies "ENGLISH" as the
                (RUN_UNITS,DEFAULT_UNITS) respectively and False, False otherwise.
                Returns (None, None) if it can't find a (RUN_UNITS, DEFAULT_UNITS) in the supplied files.
        """
        oilfield_run_units: Optional[bool] = None
        oilfield_default_units: Optional[bool] = None

        for model in models:
            # If we're checking the units of a RESQML model, read it in and get the units. Otherwise, read the units
            # from the fcs file
            if os.path.splitext(model)[1] == '.epc':
                resqpy_model = rq.Model(epc_file=model)

                # Load in the RESQML grid
                grid = resqpy_model.grid()

                # Check the grid units
                grid_length_unit = grid.xy_units()
                model_oilfield_default_units = grid_length_unit == 'ft'
                model_oilfield_run_units = grid_length_unit == 'ft'
            else:
                simulator = NexusSimulator(origin=model)
                model_oilfield_default_units = simulator.default_units == UnitSystem.ENGLISH
                model_oilfield_run_units = simulator.run_units == UnitSystem.ENGLISH

            # If not defined, assign it to model_oilfield_default_units
            if oilfield_default_units is None:
                oilfield_default_units = model_oilfield_default_units
            # Raise ValueError if default units are inconsistent with the other models
            elif model_oilfield_default_units != oilfield_default_units:
                raise ValueError(f"Model at {model} using inconsistent default units")

            # If not defined, assign it to model_oilfield_run_units
            if oilfield_run_units is None:
                oilfield_run_units = model_oilfield_run_units
            # Raise ValueError if run units are inconsistent with the other models
            elif model_oilfield_run_units != oilfield_run_units:
                raise ValueError(f"Model at {model} using inconsistent run units")

        return oilfield_run_units, oilfield_default_units

    @staticmethod
    def get_check_oil_gas_types_for_models(models: list[str]) -> Optional[str]:
        """Checks for fluid types within a list of paths to models.
        Currently limited to checking for the first SURFACE network in a file.

        Args:
            models (list[str]): a list of paths to models to check for fluid types.

        Raises:
            ValueError: If fluid types are inconsistent between models

        Returns:
            Optional[str]: The fluid type used for the model for the first surface network
        """
        fluid_type = None
        for model in models:
            model_fluid_type = None
            fcs_file = NexusFile.generate_file_include_structure(model).get_flat_list_str_file
            surface_filename = None
            if fcs_file is None:
                warnings.warn(UserWarning(f'No file found for {model}'))
                continue
            for line in fcs_file:
                if nfo.check_token("SURFACE Network 1", line):
                    surface_filename = nfo.get_expected_token_value(token="SURFACE Network 1", token_line=line,
                                                                    file_list=fcs_file)
                    break

            if surface_filename is not None:
                surface_filename = surface_filename if os.path.isabs(surface_filename) else \
                    os.path.dirname(model) + "/" + surface_filename
                model_fluid_type = NexusSimulator.get_fluid_type(
                    surface_file_name=surface_filename)

            if fluid_type is None:
                fluid_type = model_fluid_type
            elif fluid_type != model_fluid_type:
                raise ValueError(
                    f"Inconsistent Oil / Gas types: {model_fluid_type} found for {model}")

        return fluid_type

    @staticmethod
    def get_eos_details(surface_file: list[str]) -> str:
        """Gets all the information about an EOS from a Nexus model.

        Args:
            surface_file (list[str]): path to the surface file in a Nexus model

        Returns:
            str: a concatenated string of EOS components
        """
        eos_string: str = ''
        eos_found: bool = False
        for line in surface_file:
            if nfo.check_token("EOS", line):
                eos_string += line
                eos_found = True
            elif eos_found:
                eos_string += line
            if nfo.check_token("COMPONENTS", line):
                break

        return eos_string

    @staticmethod
    def get_fluid_type(surface_file_name: str) -> str:
        """Gets the fluid type for a single model from a surface file.

        Args:
            surface_file_name (str): path to the surface file in a Nexus model

        Raises:
            ValueError: if no fluid type is found within the provided file path

        Returns:
            str: fluid type as one of [BLACKOIL, WATEROIL, GASWATER, API] or the full details from an EOS model
        """
        surface_file = nfo.load_file_as_list(surface_file_name)
        fluid_type = None

        for line in surface_file:
            if nfo.check_token("BLACKOIL", line):
                fluid_type = "BLACKOIL"
                break
            elif nfo.check_token("WATEROIL", line):
                fluid_type = "WATEROIL"
                break
            elif nfo.check_token("GASWATER", line):
                fluid_type = "GASWATER"
                break
            elif nfo.check_token("EOS", line):
                fluid_type = NexusSimulator.get_eos_details(surface_file)
            elif nfo.check_token("API", line):
                fluid_type = "API"

        if fluid_type is None:
            raise ValueError("No Oil / Gas type detected")

        return fluid_type

    def get_model_oil_type(self) -> str:
        """Returns the get_fluid_type method on the existing NexusSimulator instance.

        Raises:
            ValueError: If no file path is provided for the surface file path

        Returns:
            str: fluid type as one of [BLACKOIL, WATEROIL, GASWATER,] or the full details from an EOS model
        """
        if self.model_files.surface_files is None or self.model_files.surface_files[1].location is None:
            raise ValueError("No value found for the path to the surface file")
        return NexusSimulator.get_fluid_type(self.model_files.surface_files[1].location)

    def check_output_path(self) -> None:
        """Confirms that the output path has been set (used to stop accidental writing operations in the original
        directory).

        Raises:
            ValueError: if the destination provided is set to None.
        """
        if self.__destination is None:
            raise ValueError("Destination is required for this operation. Currently set to: ", self.__destination)

    @property
    def destination(self) -> Optional[str]:
        return self.__destination

    def set_output_path(self, path: str) -> None:
        """Initialises the output to the declared output location.
        If the file is a different directory to the origin path location the function will set the origin
        to the new destination.
        """
        self.__destination = path
        if self.__destination is not None and os.path.dirname(self._origin) != os.path.dirname(self.__destination):
            self._origin = self.__destination + "/" + os.path.basename(self.__original_fcs_file_path)

    def __load_fcs_file(self):
        """Loads in the information from the supplied FCS file into the class instance.
        Loads in the paths for runcontrol, structured grid and the first surface network.
        Loads in the values for dateformat and run units.
        Attempts to load the run_control_file.
        Loads the wellspec and dynamic property files.
        """
        # fcs_content_with_includes is used to scan only the fcs file and files specifically called with the INCLUDE
        # token in front of it to prevent it from reading through all the other files. We need this here to extract the
        # fcs properties only. The FcsFile structure is then generated and stored in the object (with all the nesting of
        # the NexusFiles as self.model_files (e.g. STRUCTURED_GRID, RUNCONTROL etc)
        fcs_content_with_includes = NexusFile.generate_file_include_structure(
            self.__new_fcs_file_path).get_flat_list_str_file
        self._model_files = FcsNexusFile.generate_fcs_structure(self.__new_fcs_file_path)
        if fcs_content_with_includes is None:
            raise ValueError(f'FCS file not found, no content for {self.__new_fcs_file_path}')
        for line in fcs_content_with_includes:
            if nfo.check_token('DATEFORMAT', line) or nfo.check_token('DATE_FORMAT', line):
                format_token = 'DATEFORMAT' if nfo.check_token('DATEFORMAT', line) else 'DATE_FORMAT'
                value = nfo.get_token_value(format_token, line, fcs_content_with_includes)
                if value is not None:
                    self.date_format = DateFormat.DD_MM_YYYY if value == 'DD/MM/YYYY' else DateFormat.MM_DD_YYYY

                self._sim_controls.date_format_string = "%m/%d/%Y" if self.date_format is DateFormat.MM_DD_YYYY \
                    else "%d/%m/%Y"
            elif nfo.check_token('RUN_UNITS', line):
                value = nfo.get_token_value('RUN_UNITS', line, fcs_content_with_includes)
                if value is not None:
                    self.__run_units = UnitSystem(value.upper())
            elif nfo.check_token('DEFAULT_UNITS', line):
                value = nfo.get_token_value('DEFAULT_UNITS', line, fcs_content_with_includes)
                if value is not None:
                    self.__default_units = UnitSystem(value.upper())

        # Load in the other files

        # === Load in dynamic properties ===
        # Read in PVT properties from Nexus PVT method files
        if self.model_files.pvt_files is not None and \
                len(self.model_files.pvt_files) > 0:
            self._pvt = NexusPVTMethods(files=self.model_files.pvt_files,
                                        model_unit_system=self.default_units)

        # Read in separator properties from Nexus separator method files
        if self.model_files.separator_files is not None and \
                len(self.model_files.separator_files) > 0:
            self._separator = NexusSeparatorMethods(files=self.model_files.separator_files,
                                                    model_unit_system=self.default_units)

        # Read in water properties from Nexus water method files
        if self.model_files.water_files is not None and \
                len(self.model_files.water_files) > 0:
            self._water = NexusWaterMethods(files=self.model_files.water_files,
                                            model_unit_system=self.default_units)

        # Read in equilibration properties from Nexus equil method files
        if self.model_files.equil_files is not None and \
                len(self.model_files.equil_files) > 0:
            self._equil = NexusEquilMethods(files=self.model_files.equil_files,
                                            model_unit_system=self.default_units)

        # Read in rock properties from Nexus rock method files
        if self.model_files.rock_files is not None and \
                len(self.model_files.rock_files) > 0:
            self._rock = NexusRockMethods(files=self.model_files.rock_files,
                                          model_unit_system=self.default_units)

        # Read in relative permeability and capillary pressure properties from Nexus relperm method files
        if self.model_files.relperm_files is not None and \
                len(self.model_files.relperm_files) > 0:
            self._relperm = NexusRelPermMethods(files=self.model_files.relperm_files,
                                                model_unit_system=self.default_units)

        # Read in valve and choke properties from Nexus valve method files
        if self.model_files.valve_files is not None and \
                len(self.model_files.valve_files) > 0:
            self._valve = NexusValveMethods(files=self.model_files.valve_files,
                                            model_unit_system=self.default_units)

        # Read in aquifer properties from Nexus aquifer method files
        if self.model_files.aquifer_files is not None and \
                len(self.model_files.aquifer_files) > 0:
            self._aquifer = NexusAquiferMethods(files=self.model_files.aquifer_files,
                                                model_unit_system=self.default_units)

        # Read in hydraulics properties from Nexus hyd method files
        if self.model_files.hyd_files is not None and \
                len(self.model_files.hyd_files) > 0:
            self._hydraulics = NexusHydraulicsMethods(files=self.model_files.hyd_files,
                                                      model_unit_system=self.default_units)

        # Read in gaslift properties from Nexus gaslift method files
        if self.model_files.gas_lift_files is not None and \
                len(self.model_files.gas_lift_files) > 0:
            self._gaslift = NexusGasliftMethods(files=self.model_files.gas_lift_files,
                                                model_unit_system=self.default_units)

        # === End of dynamic properties loading ===

        # Load in Runcontrol
        if self.model_files.runcontrol_file is not None:
            self.run_control_file_path = self.model_files.runcontrol_file.location
            self._sim_controls.load_run_control_file()

        if self.model_files.structured_grid_file is not None:
            self._grid = NexusGrid.load_structured_grid_file(self.model_files.structured_grid_file,
                                                             lazy_loading=self.__lazy_loading)

        # Load in wellspec files
        if self.model_files.well_files is not None and \
                len(self.model_files.well_files) > 0:
            for well_file in self.model_files.well_files.values():
                if well_file.location is None:
                    warnings.warn(f'Well file location has not been found for {well_file}')
                    continue

    @staticmethod
    def update_file_value(file_path: str, token: str, new_value: str, add_to_start: bool = False) -> None:
        """Updates a value in a file if it is present and in the format {TOKEN} {VALUE}. If the token
        isn't present, it will add the token + value to either the start or end of the file.

        Args:
            file_path (str): path to a file to update the token/value pair in
            token (str): Keyword token to find in the given file (e.g. KX)
            new_value (str): Value following the TOKEN to be replaced
            add_to_start (bool, optional): Inserts the token/value pair to the start of the file. Defaults to False.

        Raises:
            ValueError: If no value is found after the token
        """

        file = nfo.load_file_as_list(file_path)

        line_counter = 0
        token_found = False
        for line in file:
            modified_line = line.lower().replace('/t', ' ')
            modified_line = ' '.join(modified_line.split())
            if token.lower() in modified_line:
                # We've found the token, now replace the value
                token_location = modified_line.find(token.lower())
                line_before_token_value = line[0: token_location]
                line_after_token = line[token_location:]
                current_value = nfo.get_next_value(0, [line], line_after_token[len(token) + 1:])
                if current_value is None:
                    raise ValueError(f"No value found after the supplied {token=}, \
                        please check the following line for that token: {line}")
                new_line_after = line_after_token.replace(
                    current_value, new_value, 1)
                file[line_counter] = line_before_token_value + new_line_after
                token_found = True
                break
            line_counter += 1

        if token_found is False:
            token_line = f"{token} {new_value}"
            if add_to_start:
                file.insert(0, token_line + '\n')
            else:
                file.append('\n' + token_line)

        new_file_str = "".join(file)

        with open(file_path, "w") as text_file:
            text_file.write(new_file_str)

    def update_fcs_file_value(self, token, new_value, add_to_start=False):
        """Updates a value in the FCS file."""
        self.update_file_value(self.__new_fcs_file_path, token=token, new_value=new_value, add_to_start=add_to_start)

    @staticmethod
    def comment_out_file_value(token: str, file_path: str) -> None:
        """Comments out an uncommented line containing the specified token.

        Args:
            token (str): Keyword token to find in the given file (e.g. KX)
            file_path (str): path to a file containing the token
        """
        file = nfo.load_file_as_list(file_path)

        line_counter = 0
        for line in file:
            modified_line = line.lower().replace('/t', ' ')
            modified_line = ' '.join(modified_line.split())
            # If we've found the token, and it isn't already commented, comment it out
            if token.lower() in modified_line and \
                    (modified_line.find(token.lower()) < modified_line.find("!") or modified_line.find("!") == -1):
                file[line_counter] = f"! {file[line_counter]}"
                break
            line_counter += 1

        new_file_str = "".join(file)

        with open(file_path, "w") as text_file:
            text_file.write(new_file_str)

    def get_date_format(self) -> str:
        """Returns the date format being used by the model
        formats used: ('MM/DD/YYYY', 'DD/MM/YYYY').
        """
        return self._sim_controls.get_date_format(self.date_format)

    def modify(self, operation: str, section: str, keyword: str, content: list[str]):
        """Generic modify method to modify part of the input deck. \
        Operations are dependent on the section being modified.

        Args:
            operation (str): operation to perform on the section of the input deck (e.g. 'merge')
            section (str): file type from the input deck provided (e.g. RUNCONTROL)
            keyword (str): which keyword/token to find within the deck provided (e.g. TIME)
            content (list[str]): The content to modify using the above operation, \
            represented as a list of strings with a new entry per line of the file

        Raises:
            NotImplementedError: if the functionality is not yet implemented
        """
        section = section.upper()
        keyword = keyword.upper()
        operation = operation.lower()

        if section == "RUNCONTROL":
            if keyword == "TIME":
                self._sim_controls.modify_times(content=content, operation=operation)
            else:
                raise NotImplementedError(keyword, "not yet implemented")
        else:
            raise NotImplementedError(section, "not yet implemented")

    def get_content(self, section: str, keyword: str) -> Union[list[str], None]:
        """Returns the requested input information.

        Args:
            section (str): Section to retreive information from
            keyword (str): Keyword/token to retrieve the information for

        Raises:
            NotImplementedError: if the functionality is not yet implemented

        Returns:
            Union[list[str], None]: the requested information
        """
        section = section.upper()
        keyword = keyword.upper()
        if section == "RUNCONTROL":
            if keyword == "TIME":
                return self._sim_controls.times
            else:
                raise NotImplementedError(keyword, "not yet implemented")
        else:
            raise NotImplementedError(section, "not yet implemented")

    def get_structured_grid_dict(self) -> dict[str, Any]:
        """Convert the structured grid info to a dictionary and pass it to the front end."""
        if self._grid is None:
            return {}
        return self._grid.to_dict()

    def get_abs_structured_grid_path(self, filename: str):
        """Returns the absolute path to the Structured Grid file."""
        if self.model_files.structured_grid_file is None:
            raise ValueError(
                f"No structured grid file found within simulator class: {self.model_files.structured_grid_file}")
        grid_path = self.model_files.structured_grid_file.location
        if grid_path is None:
            raise ValueError("No path found for structured grid file path. \
                Please provide a path to the structured grid")
        return os.path.dirname(grid_path) + '/' + filename

    def get_surface_file_path(self):
        """Get the surface file path."""
        if self.model_files.surface_files is None or self.model_files.surface_files[1] is None:
            raise ValueError('No path found for surface file.')
        return self.model_files.surface_files[1].location

    def load_network(self):
        """Populates nodes and connections from a surface file."""
        self._network.load()

    def write_out_case(self, new_file_path: str, new_include_file_location: str = 'include_files',
                       case_suffix: str = '') -> None:
        """Creates a new case of a simulator run by writing out the modified include files to a new file.

        Args:
            new_file_path (str): Path to save the new simulator to.
            new_include_file_location (str): Saves included files to a path either absolute or relative to the
            file path provided.
            case_suffix (str): Suffix to append to the case name. Defaults to ''.
        """

        self.model_files.write_out_case(new_file_path=new_file_path,
                                        new_include_file_location=new_include_file_location,
                                        case_suffix=case_suffix)

    def update_simulator_files(self) -> None:
        """Updates the simulator with any changes to the included files. Overwrites existing files.
        IMPORTANT: No changes to the model will be saved until this method is called!
        """

        self.model_files.update_model_files()

    def move_simulator_files(self, new_file_path: str, new_include_file_location: str) -> None:
        """Creates a set of simulator files.

        Args:
            new_file_path (str): Path to save the new simulator to.
            new_include_file_location (str): Saves included files to a path either absolute or relative to the
            file path provided.
        """

        self.model_files.move_model_files(new_file_path, new_include_file_location)

    def write_out_new_model(self, new_location: str, new_model_name: str) -> None:
        """Not implemented for Nexus yet."""
        raise NotImplementedError("Not Implemented Yet")



from __future__ import annotations
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Optional
from uuid import UUID


@dataclass
class FileBase(ABC):
    """The abstract base class for simulator files.

    Attributes:
        location (str): Full path to file location
        file_content_as_list (list[str]): List of lines in the file
    """

    location: str
    file_content_as_list: Optional[list[str]] = field(default=None, repr=False)
    __file_modified: bool = False

    @abstractmethod
    def write_to_file(self) -> None:
        raise NotImplementedError("Implement this in the derived class")

    @property
    @abstractmethod
    def get_flat_list_str_file(self) -> list[str]:
        raise NotImplementedError("Implement this in the derived class")

    @abstractmethod
    def add_object_locations(self, obj_uuid: UUID, line_indices: list[int]) -> None:
        raise NotImplementedError("Implement this in the derived class")

    @staticmethod
    @abstractmethod
    def insert_comments(additional_content: list[str], comments) -> list[str]:
        raise NotImplementedError("Implement this in the derived class")

    @abstractmethod
    def get_object_locations_for_id(self, obj_id: UUID) -> list[int]:
        raise NotImplementedError("Implement this in the derived class")

    @abstractmethod
    def remove_object_from_file_as_list(self, objects_to_remove: list[UUID]) -> None:
        raise NotImplementedError("Implement this in the derived class")

    @abstractmethod
    def add_to_file_as_list(self, additional_content: list[str], index: int,
                            additional_objects: Optional[dict[UUID, list[int]]] = None,
                            comments: Optional[str] = None) -> None:
        raise NotImplementedError("Implement this in the derived class")

    @abstractmethod
    def remove_from_file_as_list(self, index: int, objects_to_remove: Optional[list[UUID]] = None,
                                 string_to_remove: Optional[str] = None) -> None:
        raise NotImplementedError("Implement this in the derived class")



from __future__ import annotations
import uuid
from abc import ABC
from dataclasses import dataclass, field
from typing import Optional
from ResSimpy.DataObjectMixin import DataObjectMixin
from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.Units.AttributeMappings.NetworkUnitMapping import NetworkUnits


@dataclass(repr=False)
class Target(DataObjectMixin, ABC):
    name: Optional[str] = None
    date: Optional[str] = None
    unit_system: Optional[UnitSystem] = None

    control_quantity: Optional[str] = None
    control_conditions: Optional[str] = None
    control_connections: Optional[str] = None
    control_method: Optional[str] = None
    calculation_method: Optional[str] = None
    calculation_conditions: Optional[str] = None
    calculation_connections: Optional[str] = None
    value: Optional[float] = None
    add_value: Optional[float] = None
    region: Optional[str] = None
    priority: Optional[int] = None
    minimum_rate: Optional[str] = None
    minimum_rate_no_shut: Optional[float] = None
    guide_rate: Optional[str] = None
    max_change_pressure: Optional[float] = None
    rank_dt: Optional[float] = None
    control_type: Optional[str] = None
    calculation_type: Optional[str] = None
    __id: uuid.UUID = field(default_factory=lambda: uuid.uuid4(), compare=False)

    @property
    def units(self) -> NetworkUnits:
        """Returns the attribute to unit map for the WellConnection."""
        return NetworkUnits(self.unit_system)



from dataclasses import dataclass, field
from abc import ABC
from typing import Literal

from ResSimpy.Node import Node
from ResSimpy.OperationsMixin import NetworkOperationsMixIn


@dataclass(kw_only=True)
class Nodes(NetworkOperationsMixIn, ABC):
    __nodes: list[Node] = field(default_factory=list)

    @property
    def _network_element_name(self) -> Literal['nodes']:
        return 'nodes'



"""The abstract base class for all simulators."""
from __future__ import annotations
import os
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Optional

from ResSimpy.Aquifer import Aquifer
from ResSimpy.Equilibration import Equilibration
from ResSimpy.File import File
from ResSimpy.Gaslift import Gaslift
from ResSimpy.Grid import Grid
from ResSimpy.Hydraulics import Hydraulics
from ResSimpy.Network import Network
from ResSimpy.PVT import PVT
from ResSimpy.RelPerm import RelPerm
from ResSimpy.Rock import Rock
from ResSimpy.Separator import Separator
from ResSimpy.Valve import Valve
from ResSimpy.Water import Water
from ResSimpy.Wells import Wells


@dataclass(kw_only=True, init=False)
class Simulator(ABC):
    _start_date: str
    _origin: str
    _wells: Wells
    _pvt: PVT
    _separator: Separator
    _water: Water
    _equil: Equilibration
    _rock: Rock
    _relperm: RelPerm
    _valve: Valve
    _aquifer: Aquifer
    _hydraulics: Hydraulics
    _gaslift: Gaslift
    _network: Network
    _grid: Optional[Grid]
    _model_files: File

    """Class Properties"""

    @property
    def start_date(self) -> str:
        return self._start_date

    @start_date.setter
    def start_date(self, value) -> None:
        self._start_date = value

    @property
    def wells(self) -> Wells:
        return self._wells

    @property
    def pvt(self) -> PVT:
        return self._pvt

    @property
    def separator(self) -> Separator:
        return self._separator

    @property
    def water(self) -> Water:
        return self._water

    @property
    def equil(self) -> Equilibration:
        return self._equil

    @property
    def rock(self) -> Rock:
        return self._rock

    @property
    def relperm(self) -> RelPerm:
        return self._relperm

    @property
    def valve(self) -> Valve:
        return self._valve

    @property
    def aquifer(self) -> Aquifer:
        return self._aquifer

    @property
    def hydraulics(self) -> Hydraulics:
        return self._hydraulics

    @property
    def gaslift(self) -> Gaslift:
        return self._gaslift

    @property
    def network(self) -> Network:
        return self._network

    @property
    def grid(self) -> Optional[Grid]:
        """Pass the grid information to the front end."""
        return self._grid

    @property
    def origin(self) -> str:
        return self._origin

    @origin.setter
    def origin(self, value: Optional[str]) -> None:
        if value is None:
            raise ValueError(f'Origin path to model is required. Instead got {value}.')
        self._origin: str = value.strip()

    @property
    def model_location(self) -> str:
        """Returns the location of the model."""
        return os.path.dirname(self._origin)

    @property
    def model_files(self) -> File:
        return self._model_files

    """ Class Methods """

    @staticmethod
    @abstractmethod
    def get_fluid_type(surface_file_name: str) -> str:
        raise NotImplementedError("This method has not been implemented for this simulator yet")

    @abstractmethod
    def set_output_path(self, path: str) -> None:
        raise NotImplementedError("Implement this method on the derived class")

    @abstractmethod
    def get_date_format(self) -> str:
        """Returns date format as a string."""
        raise NotImplementedError("Implement this method on the derived class")

    @abstractmethod
    def write_out_new_model(self, new_location: str, new_model_name: str) -> None:
        """Writes out a new version of the model to the location supplied.

        Args:
        new_location (str): Path to write the contents of the model to.
        """
        raise NotImplementedError("Implement this method on the derived class")



from __future__ import annotations
from dataclasses import dataclass
import os
from typing import Optional, MutableMapping
from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.Nexus.DataModels.NexusFile import NexusFile
from ResSimpy.Nexus.DataModels.NexusGasliftMethod import NexusGasliftMethod
from ResSimpy.Gaslift import Gaslift


@dataclass(kw_only=True)
class NexusGasliftMethods(Gaslift):
    """Class for collection of Nexus gaslift methods.

    Attributes:
        inputs (dict[int, NexusGasliftMethod]): Collection of Nexus gaslift methods, as a dictionary
        files (dict[int, NexusFile]): Dictionary collection of gaslift files, as defined in Nexus fcs file.
    """

    __inputs: MutableMapping[int, NexusGasliftMethod]
    __files: dict[int, NexusFile]
    __properties_loaded: bool = False  # Used in lazy loading
    __model_unit_system: UnitSystem

    def __init__(self, model_unit_system: UnitSystem,
                 inputs: Optional[MutableMapping[int, NexusGasliftMethod]] = None,
                 files: Optional[dict[int, NexusFile]] = None) -> None:
        if inputs:
            self.__inputs = inputs
        else:
            self.__inputs: MutableMapping[int, NexusGasliftMethod] = {}
        if files:
            self.__files = files
        else:
            self.__files = {}
        self.__model_unit_system = model_unit_system
        super().__init__()

    def __repr__(self) -> str:
        """Pretty printing gaslift methods."""
        if not self.__properties_loaded:
            self.load_gaslift_methods()
        printable_str = ''
        for table_num in self.__inputs.keys():
            printable_str += '\n--------------------------------\n'
            printable_str += f'GASLIFT method {table_num}\n'
            printable_str += '--------------------------------\n'
            printable_str += self.__inputs[table_num].__repr__()
            printable_str += '\n'

        return printable_str

    @property
    def inputs(self) -> MutableMapping[int, NexusGasliftMethod]:
        if not self.__properties_loaded:
            self.load_gaslift_methods()
        return self.__inputs

    @property
    def files(self) -> dict[int, NexusFile]:
        return self.__files

    def load_gaslift_methods(self):
        # Read in gaslift properties from Nexus gaslift method files
        if self.__files is not None and len(self.__files) > 0:  # Check if gaslift files exist
            for table_num in self.__files.keys():  # For each gaslift property method
                gaslift_file = self.__files[table_num]
                if gaslift_file.location is None:
                    raise ValueError(f'Unable to find gaslift file: {gaslift_file}')
                if os.path.isfile(gaslift_file.location):
                    # Create NexusGasliftMethod object
                    self.__inputs[table_num] = NexusGasliftMethod(file=gaslift_file, input_number=table_num,
                                                                  model_unit_system=self.__model_unit_system)
                    self.__inputs[table_num].read_properties()  # Populate object with gaslift props in file
        self.__properties_loaded = True



"""Base class representing a data object in ResSimpy."""
import uuid
from abc import ABC, abstractmethod
from dataclasses import dataclass, field

from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.Units.AttributeMappings.BaseUnitMapping import BaseUnitMapping
from ResSimpy.Utils import to_dict_generic
from ResSimpy.Utils.generic_repr import generic_repr, generic_str
from ResSimpy.Utils.obj_to_table_string import to_table_line


@dataclass()
class DataObjectMixin(ABC):
    """Base class representing a data object in ResSimpy."""
    __id: uuid.UUID = field(default_factory=lambda: uuid.uuid4(), compare=False, repr=False)

    def __init__(self, properties_dict: dict[str, None | int | str | float]) -> None:
        # properties dict is a parameter to make the call signature equivalent to subclasses.
        self.__id = uuid.uuid4()
        if properties_dict:
            raise ValueError('No properties should be passed to the DataObjectMixin')

    def __repr__(self) -> str:
        return generic_repr(self)

    def __str__(self) -> str:
        return generic_str(self)

    def to_dict(self, keys_in_keyword_style: bool = False, add_date: bool = True, add_units: bool = True,
                include_nones: bool = True) -> dict[str, None | str | int | float]:
        """Returns a dictionary of the attributes of the object.

        Args:
            include_nones (bool):
            keys_in_keyword_style (bool): if True returns the key values as simulator keywords, otherwise returns the \
                attribute name as stored by ressimpy.
            add_date (bool): if True adds the date to the dictionary
            add_units (bool): if True adds the units to the dictionary

        Returns:
            a dictionary keyed by attributes and values as the value of the attribute
        """
        result_dict = to_dict_generic.to_dict(self, keys_in_keyword_style, add_date=add_date, add_units=add_units,
                                              include_nones=include_nones)
        return result_dict

    def to_table_line(self, headers: list[str]) -> str:
        """Takes a generic Nexus object and returns the attribute values as a string in the order of headers provided.
        Requires an implemented to_dict method and get_keyword_mapping() method.

        Args:
            headers (list[str]): list of header values in keyword format

        Returns:
            string of the values in the order of the headers provided.

        """
        return to_table_line(self, headers)

    @property
    def id(self) -> uuid.UUID:
        """Unique identifier for each object."""
        return self.__id

    @staticmethod
    @abstractmethod
    def get_keyword_mapping() -> dict[str, tuple[str, type]]:
        """Gets the mapping of keywords to attribute definitions."""
        raise NotImplementedError("Implement this in the derived class")

    @property
    @abstractmethod
    def units(self) -> BaseUnitMapping:
        """Returns the attribute to unit map for the data object."""
        raise NotImplementedError("Implement this in the derived class")

    def get_unit_for_attribute(self, attribute_name: str, unit_system: UnitSystem, uppercase: bool = False) -> str:
        """Returns the unit variable for the given unit system.

        Args:
            attribute_name (str): name of the attribute to get the unit for
            unit_system (UnitSystem): unit system to get the unit for
            uppercase (bool): if True returns the unit in uppercase
        """
        unit_dimension = self.units.attribute_map.get(attribute_name, None)
        if unit_dimension is None:
            raise AttributeError(f'Attribute {attribute_name} not recognised and does not have a unit definition')
        unit = unit_dimension.unit_system_enum_to_variable(unit_system=unit_system)
        if uppercase:
            unit = unit.upper()
        return unit



"""A collection of functions for handling grid functions from Nexus."""
from __future__ import annotations
import ResSimpy.Nexus.nexus_file_operations as nfo
import pandas as pd
from typing import Union
import warnings
from ResSimpy.Nexus.NexusKeywords.structured_grid_keywords import GRID_ARRAY_KEYWORDS


def collect_all_function_blocks(file_as_list: list[str]) -> list[list[str]]:
    """Collects all the function blocks within a grid file.

    Args:
    file_as_list (list[str] | NexusFile): a list of strings containing each line of the file as an item,

    Returns:
    list[list[str]]: list of function block lines as a list of strings.
    """
    function_list = []
    function_body: list[str] = []
    reading_function = False

    for i, line in enumerate(file_as_list):
        if nfo.check_token('FUNCTION', line):
            function_body = []
            reading_function = True
        if reading_function:
            # remove all comments following the first '!' in a line.
            modified_line = line.split('!', 1)[0]
            function_body.append(modified_line.strip())
            if nfo.check_token('OUTPUT', modified_line) and not nfo.check_token('RANGE', modified_line):
                function_list.append(function_body)
                reading_function = False
    # remove null values
    function_list = [list(filter(None, x)) for x in function_list]

    return function_list


def create_function_parameters_df(function_list_to_parse: list[list[str]]) -> pd.DataFrame:
    """Creates a dataframe to hold all the function properties and parameters.

    Args:
        function_list_to_parse (list[list[str]]): list of functions extracted as list of lines.

    Returns:
        pandas.DataFrame: a dataframe holding each function's parameters in a row.
    """

    functions_df = pd.DataFrame(
        columns=['FUNCTION #', 'blocks [i1,i2,j1,j2,k1,k2]', 'region_type',
                 'region_numbers',
                 'func_type', 'func_coeff', 'grid', 'range_input', 'range_output', 'drange',
                 'input_arrays', 'output_arrays', 'i1', 'i2', 'j1', 'j2', 'k1', 'k2'])

    for b, block in enumerate(function_list_to_parse):

        # set the empty default values for the parameters,
        # so if they don't exist in dataframe they won't appear as NaN or give error,
        # or repeat the last value for each row.
        i1 = i2 = j1 = j2 = k1 = k2 = region_type = function_type = grid_name = ''
        # set the lists as empty strings as well, otherwise they show up as [] on the dataframe.
        region_number_list: Union[str, list[str], list[int]] = ''
        function_coefficients: Union[str, list[str], list[float]] = ''
        input_arrays_min_max_list: Union[str, list[str], list[float]] = ''
        output_arrays_min_max_list: Union[str, list[str], list[float]] = ''
        input_array_list: Union[str, list[str]] = ''
        output_array_list: Union[str, list[str]] = ''
        drange_list: Union[str, list[str]] = ''
        blocks_list: Union[str, list[str], list[int]] = ''

        for li, line in enumerate(block):
            modified_line = line.upper()
            words = modified_line.split()
            if 'BLOCKS' in modified_line:
                i1 = words[1]
                i2 = words[2]
                j1 = words[3]
                j2 = words[4]
                k1 = words[5]
                k2 = words[6]
                blocks_list = words[1:7]
                blocks_list = [round(float(i)) for i in blocks_list]

            if 'FUNCTION' in modified_line:
                if len(words) == 1:
                    continue
                if len(words) == 2:
                    if words[1] not in GRID_ARRAY_KEYWORDS:
                        warnings.warn(f'Function {b + 1}:  Function table entries will be excluded from summary df.')
                        function_type = 'function table'
                    else:
                        region_type = words[1]
                        region_number_list = block[li + 1].split()
                        try:
                            region_number_list = [round(float(i)) for i in region_number_list]
                        except ValueError:
                            warnings.warn(f'ValueError at function {b + 1}: could not convert string to integer.')

                if len(words) > 2:  # TODO: deal with tabular function option keywords
                    warnings.warn(f'Function {b + 1}:  Function table entries will be excluded from summary df.')
                    function_type = 'function table'
            if 'ANALYT' in modified_line:
                function_type = words[1]
                if len(words) > 2:
                    # remove the first 2 words in line, and set the rest to coefficients
                    function_coefficients = words[2:]
                    # convert string coefficient values to numerical, if possible:
                    try:
                        function_coefficients = [float(i) for i in function_coefficients]
                    except ValueError:
                        warnings.warn(f'ValueError at function {b + 1}: could not convert string to float.')

            if 'GRID' in modified_line:
                grid_name = words[1]
            if 'RANGE' in modified_line and 'INPUT' in modified_line:
                input_arrays_min_max_list = words[2:]
                # convert string range_input values to numerical, if possible:
                try:
                    input_arrays_min_max_list = [float(i) for i in input_arrays_min_max_list]
                except ValueError:
                    warnings.warn(f'ValueError at function {b + 1}: could not convert string to float.')
            if 'RANGE' in modified_line and 'OUTPUT' in modified_line:
                output_arrays_min_max_list = words[2:]
                # convert string range_input values to numerical, if possible:
                try:
                    output_arrays_min_max_list = [float(i) for i in output_arrays_min_max_list]
                except ValueError:
                    warnings.warn(f'ValueError at function {b + 1}: could not convert string to float.')
            if 'DRANGE' in modified_line:
                warnings.warn(f'Function {b + 1}: Function table entries will be excluded from summary df.')
                drange_list = words[1:]
                function_type = 'function table'
            if 'OUTPUT' in modified_line and 'RANGE' not in modified_line:
                input_array_list = words[:words.index('OUTPUT')]
                output_array_list = words[words.index('OUTPUT') + 1:]
        # Create the row that holds function data
        function_row = [b + 1, blocks_list, region_type, region_number_list, function_type,
                        function_coefficients,
                        grid_name, input_arrays_min_max_list, output_arrays_min_max_list, drange_list, input_array_list,
                        output_array_list, i1, i2, j1, j2, k1, k2]
        # Append the function row to the dataframe
        functions_df.loc[len(functions_df)] = function_row
    return functions_df


def summarize_model_functions(function_list_to_parse: list[list[str]]) -> pd.DataFrame:
    """Extracts all function parameters into a df, with an added column of human-readable notations for each function.

    Args:
        function_list_to_parse (list[list[str]]): list of functions extracted as list of lines.

    Returns:
        pandas.DataFrame: a dataframe holding each function's translation/summary in a row.
    """

    # get the df from create_function_parameters_df, add a new column, and populate based on ANALYT function type:

    function_summary_df = create_function_parameters_df(function_list_to_parse)
    function_summary_df['notation'] = ''

    for index, row in function_summary_df.iterrows():

        formula = row['output_arrays'][0] + ' = '

        match row['func_type'].upper():
            # ANALYT POLYN
            case 'POLYN':
                # get number of coefficients, n
                n = len(row['func_coeff'])

                # For each coefficient (c), calculate the corresponding exponent -or power (p)
                # Create polynomial function notation, term by term, for any number of coefficients (n)
                # using the first item in input arrays list (arr),
                # and the position (x) of the function term/portion we are creating.
                for x in range(n):
                    c = row['func_coeff'][x]
                    p = n - x - 1
                    arr = row['input_arrays'][0]

                    if n == 1:
                        f_portion = f'{c}'

                    if p == 0 and x > 0:
                        if c > 0:
                            f_portion = f' +{c}'
                        elif c < 0:
                            f_portion = f'{c}'
                        else:
                            continue

                    if p == 1 and x > 0:
                        if c > 0:
                            f_portion = f' +{c}*{arr}'
                        elif c < 0:
                            f_portion = f'{c}*{arr}'
                        else:
                            continue

                    if p == 1 and x == 0:
                        if c != 0:
                            f_portion = f'{c}*{arr}'
                        else:
                            continue

                    if p > 1 and x > 0:
                        if c > 0:
                            f_portion = f' +{c}*({arr}^{p})'
                        elif c < 0:
                            f_portion = f'{c}*({arr}^{p})'
                        else:
                            continue

                    if p > 1 and x == 0:
                        if c != 0:
                            f_portion = f'{c}*({arr}^{p})'
                        else:
                            continue

                    formula += f_portion

            # ANALYT ABS
            case 'ABS':
                formula += f"| {row['input_arrays'][0]} |"

            # ANALYT EXP
            case 'EXP':
                formula += f"e^{row['input_arrays'][0]}"

            # ANALYT EXP10
            case 'EXP10':
                formula += f"10^{row['input_arrays'][0]}"

            # ANALYT LOG
            case 'LOG':
                formula += f"ln|{row['input_arrays'][0]}|"

            # ANALYT LOG10
            case 'LOG10':
                formula += f"log10|{row['input_arrays'][0]}|"

            # ANALYT SQRT
            case 'SQRT':
                formula += f"SQRT|{row['input_arrays'][0]}|"

            # ANALYT GE
            case 'GE':
                formula += f"({row['input_arrays'][0]} if {row['input_arrays'][0]} >= {row['func_coeff'][0]}; " \
                           f"{row['func_coeff'][1]} otherwise)"

            # ANALYT LE
            case 'LE':
                formula += f"({row['input_arrays'][0]} if {row['input_arrays'][0]} <= {row['func_coeff'][0]}; " \
                           f"{row['func_coeff'][1]} otherwise)"

            # ANALYT ADD
            case 'ADD':
                formula += f"{row['input_arrays'][0]} + {row['input_arrays'][1]}"

            # ANALYT SUBT
            case 'SUBT':
                formula += f"{row['input_arrays'][0]} - {row['input_arrays'][1]}"

            # ANALYT DIV
            case 'DIV':
                formula += f"({row['input_arrays'][0]} / {row['input_arrays'][1]} if {row['input_arrays'][1]} != 0; " \
                           f"{row['input_arrays'][0]} otherwise)"

            # ANALYT MULT
            case 'MULT':
                formula += f"{row['input_arrays'][0]} * {row['input_arrays'][1]}"

            # ANALYT MIN
            case 'MIN':
                formula += f"min({row['input_arrays'][0]}, {row['input_arrays'][1]})"

            # ANALYT MAX
            case 'MAX':
                formula += f"max({row['input_arrays'][0]}, {row['input_arrays'][1]})"

        # fill in the notation value for the row
        function_summary_df.loc[index, 'notation'] = formula

    # move notation column to the beginning of table:
    second_column = function_summary_df.pop('notation')
    function_summary_df.insert(1, 'notation', second_column)

    # set FUNCTION number as the index colum:
    function_summary_df = function_summary_df.set_index('FUNCTION #')

    return function_summary_df



from dataclasses import dataclass, field
from abc import ABC
from typing import Literal

from ResSimpy.Target import Target
from ResSimpy.OperationsMixin import NetworkOperationsMixIn


@dataclass(kw_only=True)
class Targets(NetworkOperationsMixIn, ABC):
    __targets: list[Target] = field(default_factory=list)

    @property
    def _network_element_name(self) -> Literal['targets']:
        return 'targets'



"""Base class for handling any dynamic property simulator inputs, for use in inputs such as PVT, relperm, etc."""
from __future__ import annotations
from abc import ABC
from dataclasses import dataclass
from typing import Optional
from ResSimpy.File import File
from ResSimpy.Units.AttributeMappings.BaseUnitMapping import BaseUnitMapping


@dataclass
class DynamicProperty(ABC):
    """The abstract base class for dynamic property simulator inputs, for use in inputs such as PVT, relperm, etc.

    Attributes:
        input_number (int): Method, table or input number, in order as entered in the simulation input deck.
    """

    input_number: int
    file: File

    def __init__(self, input_number: int, file: File) -> None:
        self.input_number: int = input_number
        self.file: File = file

    @property
    def units(self) -> BaseUnitMapping:
        """Returns the attribute to unit map for the constraint."""
        raise NotImplementedError('Implement in the derived class.')

    def __repr__(self) -> str:
        """Pretty printing dynamic property data."""
        printable_str = f'\nFILE_PATH: {self.file.location}\n\n'
        printable_str += self.to_string()
        return printable_str

    def to_string(self) -> str:
        """Write dynamic property data to string."""
        raise NotImplementedError('Implement in the derived class.')

    def write_to_file(self, overwrite_existing: bool = False, new_file_location: Optional[str] = None) -> None:
        """Write dynamic property data to file."""
        printable_str = self.to_string()
        new_file_contents = printable_str.splitlines(keepends=True)
        if overwrite_existing is True and new_file_location is not None:
            raise ValueError('Please specify only one of either overwrite_existing or new_file_location.')

        if new_file_location is not None:
            new_file = File(file_content_as_list=new_file_contents, location=new_file_location)
            new_file.write_to_file()
            return
        elif overwrite_existing is False:
            raise ValueError('Please specify either overwrite_existing as True or provide new_file_location.')

        # Overwriting existing file contents
        if overwrite_existing is True:
            self.file.file_content_as_list = new_file_contents
            self.file.write_to_file()



"""Class to hold Nexus Aquifer properties."""
from __future__ import annotations
from dataclasses import dataclass, field
from enum import Enum
from typing import Optional, Union
import pandas as pd
from ResSimpy.Nexus.DataModels.NexusFile import NexusFile
from ResSimpy.Nexus.NexusKeywords.aquifer_keywords import AQUIFER_SINGLE_KEYWORDS, AQUIFER_TABLE_KEYWORDS
from ResSimpy.Nexus.NexusKeywords.aquifer_keywords import AQUIFER_KEYWORDS_VALUE_FLOAT, AQUIFER_KEYWORDS_VALUE_INT
from ResSimpy.Nexus.NexusKeywords.aquifer_keywords import AQUIFER_KEYWORDS, AQUIFER_TYPE_KEYWORDS
from ResSimpy.Enums.UnitsEnum import UnitSystem, SUnits, TemperatureUnits
from ResSimpy.DynamicProperty import DynamicProperty
from ResSimpy.Units.AttributeMappings.DynamicPropertyUnitMapping import AquiferUnits

from ResSimpy.Utils.factory_methods import get_empty_dict_union
import ResSimpy.Nexus.nexus_file_operations as nfo


@dataclass(kw_only=True, repr=False)  # Doesn't need to write an _init_, _eq_ methods, etc.
class NexusAquiferMethod(DynamicProperty):
    """Class to hold Nexus Aquifer properties.

    Attributes:
        file (NexusFile): Nexus aquifer properties file object
        input_number (int): Aquifer properties method number in Nexus fcs file
        properties (dict[str, Union[str, int, float, Enum, list[str], pd.DataFrame,
                    dict[str, Union[float, pd.DataFrame]]]]):
            Dictionary holding all properties for a specific aquifer properties method. Defaults to empty dictionary.
    """

    # General parameters
    file: NexusFile
    properties: dict[str, Union[str, int, float, Enum, list[str], pd.DataFrame,
                                dict[str, Union[float, pd.DataFrame]]]] = field(default_factory=get_empty_dict_union)
    unit_system: UnitSystem

    def __init__(self, file: NexusFile, input_number: int, model_unit_system: UnitSystem,
                 properties: Optional[dict[str, Union[str, int, float, Enum, list[str], pd.DataFrame,
                                      dict[str, Union[float, pd.DataFrame]]]]] = None) -> None:
        if properties is not None:
            self.properties = properties
        else:
            self.properties = {}
        self.unit_system = model_unit_system
        super().__init__(input_number=input_number, file=file)

    @staticmethod
    def get_keyword_mapping() -> dict[str, tuple[str, type]]:
        """Gets the mapping of nexus keywords to attribute definitions."""
        keywords: dict[str, tuple[str, type]] = {
            'BAQ': ('carter_tracy_constant', float),
            'CT': ('total_compressibility', float),
            'POR': ('porosity', float),
            'H': ('thickness', float),
            'RO': ('radius_to_inner_perimeter', float),
            'RE': ('radius_to_exterior_perimeter', float),
            'S': ('fraction_of_circular_boundary', float),
            'W': ('linear_aquifer_width', float),
            'L': ('linear_aquifer_length', float),
            'TC': ('time_conversion_factor', float),
            'VISC': ('viscosity', float),
            'K': ('permeability', float),
            'PAQI': ('initial_aquifer_pressure', float),
            'DAQI': ('datum_depth', float),
            'PI': ('productivity_index', float),
            'WEI': ('initial_encroachable_water_volume', float),
            'WAQI': ('initial_aquifer_volume', float),
        }
        return keywords

    @property
    def units(self) -> AquiferUnits:
        """Returns the attribute to unit map for the Aquifer method."""
        # Specify unit system, if provided
        if 'UNIT_SYSTEM' in self.properties.keys() and isinstance(self.properties['UNIT_SYSTEM'], UnitSystem):
            self.unit_system = self.properties['UNIT_SYSTEM']
        return AquiferUnits(unit_system=self.unit_system)

    def to_string(self) -> str:
        """Create string with aquifer data in Nexus file format."""
        printable_str = ''
        aquifer_dict = self.properties
        for key, value in aquifer_dict.items():
            if isinstance(value, pd.DataFrame):
                printable_str += f'{key}\n'
                printable_str += value.to_string(na_rep='', index=False) + '\n'
                if key == 'TRACER':
                    printable_str += 'ENDTRACER\n'
                printable_str += '\n'
            elif isinstance(value, Enum):
                if isinstance(value, UnitSystem) or isinstance(value, TemperatureUnits):
                    printable_str += f'{value.value}\n'
                elif isinstance(value, SUnits):
                    printable_str += f'SUNITS {value.value}\n'
            elif key == 'DESC' and isinstance(value, list):
                for desc_line in value:
                    printable_str += 'DESC ' + desc_line + '\n'
            elif key not in ['DESC']:
                if value == '':
                    printable_str += f'{key}\n'
                else:
                    printable_str += f'{key} {value}\n'
        printable_str += '\n'
        return printable_str

    def read_properties(self) -> None:
        """Read Nexus aquifer file contents and populate the NexusValveMethod object."""
        file_as_list = self.file.get_flat_list_str_file

        # Check for common input data
        nfo.check_for_and_populate_common_input_data(file_as_list, self.properties)

        # Specify unit system, if provided
        if 'UNIT_SYSTEM' in self.properties.keys() and isinstance(self.properties['UNIT_SYSTEM'], UnitSystem):
            self.unit_system = self.properties['UNIT_SYSTEM']

        # Initialize flags and containers to use to record properties as we iterate through aquifer file contents
        # Dictionary to record start and ending indices for tables
        aquifer_table_indices: dict[str, list[int]] = {}
        # Dictionary of flags indicating which tables are being read
        table_being_read: dict[str, bool] = {}
        for table_name in AQUIFER_TABLE_KEYWORDS:
            table_being_read[table_name] = False

        line_indx = 0
        for line in file_as_list:

            # Find standalone aquifer keywords, such as CARTER_TRACY or LINEAR
            if [i for i in line.split() if i in AQUIFER_TYPE_KEYWORDS + AQUIFER_SINGLE_KEYWORDS]:
                for word in AQUIFER_TYPE_KEYWORDS + AQUIFER_SINGLE_KEYWORDS:
                    if nfo.check_token(word, line):
                        self.properties[word] = ''
            # Find AQUIFER key-float value pairs, such as LINFAC 0.8 or BAQ 20.
            if [i for i in line.split() if i in AQUIFER_KEYWORDS_VALUE_FLOAT]:
                for key in AQUIFER_KEYWORDS_VALUE_FLOAT:
                    if nfo.check_token(key, line):
                        self.properties[key] = float(nfo.get_expected_token_value(key, line, file_as_list))
            # Find AQUIFER key-int value pairs, such as ITDPD 1 or IWATER 2
            if [i for i in line.split() if i in AQUIFER_KEYWORDS_VALUE_INT]:
                for key in AQUIFER_KEYWORDS_VALUE_INT:
                    if nfo.check_token(key, line):
                        self.properties[key] = int(nfo.get_expected_token_value(key, line, file_as_list))

            # Find beginning and ending indices of tables
            for table_key in AQUIFER_TABLE_KEYWORDS:
                if (table_key == 'TRACER' and table_being_read[table_key] and nfo.check_token('END' + table_key,
                                                                                              line)) \
                        or (table_key == 'TDPD' and table_being_read[table_key] and
                            [i for i in line.split() if nfo.check_token(i, line) and i in AQUIFER_KEYWORDS]):
                    table_being_read[table_key] = False
                    aquifer_table_indices[table_key][1] = line_indx
                if nfo.check_token(table_key, line):
                    table_being_read[table_key] = True
                    aquifer_table_indices[table_key] = [line_indx + 1, len(file_as_list)]

            line_indx += 1

        # Read in tables
        for key in aquifer_table_indices.keys():
            self.properties[key] = nfo.read_table_to_df(file_as_list[
                                                        aquifer_table_indices[key][0]:aquifer_table_indices[key][1]])



"""Class for handling wellheads in the Nexus Network. This class is used to store and manipulate the wellheads in \
a NexusNetwork. It is stored as an instance in the NexusNetwork class as "wellheads".
In Nexus this is the WELLHEAD table.
"""
from __future__ import annotations
from dataclasses import dataclass, field
from typing import Optional, TYPE_CHECKING
from uuid import UUID

import pandas as pd

from ResSimpy.File import File
from ResSimpy.Nexus.nexus_add_new_object_to_file import AddObjectOperations
from ResSimpy.Nexus.nexus_collect_tables import collect_all_tables_to_objects
from ResSimpy.Nexus.DataModels.Network.NexusWellhead import NexusWellhead
from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.Nexus.nexus_modify_object_in_file import ModifyObjectOperations
from ResSimpy.Nexus.nexus_remove_object_from_file import RemoveObjectOperations
from ResSimpy.Utils.obj_to_dataframe import obj_to_dataframe
from ResSimpy.Wellheads import Wellheads

if TYPE_CHECKING:
    from ResSimpy.Nexus.NexusNetwork import NexusNetwork


@dataclass
class NexusWellheads(Wellheads):
    __wellheads: list[NexusWellhead] = field(default_factory=list)

    def __init__(self, parent_network: NexusNetwork) -> None:
        self.__parent_network: NexusNetwork = parent_network
        self.__wellheads: list[NexusWellhead] = []
        self.__add_object_operations = AddObjectOperations(NexusWellhead, self.table_header, self.table_footer,
                                                           self.__parent_network.model)
        self.__remove_object_operations = RemoveObjectOperations(self.__parent_network, self.table_header,
                                                                 self.table_footer)
        self.__modify_object_operations = ModifyObjectOperations(self)

    def get_all(self) -> list[NexusWellhead]:
        """Returns a list of wellheads loaded from the simulator."""
        self.__parent_network.get_load_status()
        return self.__wellheads

    def get_by_name(self, name: str) -> Optional[NexusWellhead]:
        """Returns a single wellhead with the provided name loaded from the simulator.

        Args:
            name (str): name of the requested well connection
        Returns:
            NexusWellhead: which has the same name as requested
        """
        to_return = filter(lambda x: False if x.name is None else x.name.upper() == name.upper(),
                           self.__wellheads)
        return next(to_return, None)

    def get_df(self) -> pd.DataFrame:
        return obj_to_dataframe(self.__wellheads)

    def get_overview(self) -> str:
        raise NotImplementedError('To be implemented')

    def load(self, surface_file: File, start_date: str, default_units: UnitSystem) -> None:
        new_wellheads = collect_all_tables_to_objects(surface_file, {'WELLHEAD': NexusWellhead},
                                                      start_date=start_date,
                                                      default_units=default_units)
        wellheads_list = new_wellheads.get('WELLHEAD')
        if isinstance(wellheads_list, dict):
            raise ValueError('Incompatible data format for additional wells. Expected type "list" instead got "dict"')
        self._add_to_memory(wellheads_list)

    def _add_to_memory(self, additional_list: Optional[list[NexusWellhead]]) -> None:
        """Extends the wellhead object by a list of wellheads provided to it.

        Args:
            additional_list (list[NexusWellhead]): list of nexus wellheads to add to the wellhead list.
        """
        if additional_list is None:
            return
        self.__wellheads.extend(additional_list)

    def remove(self, obj_to_remove: dict[str, None | str | float | int] | UUID) -> None:
        """Remove a wellhead from the network based on the properties matching a dictionary or id.

        Args:
            obj_to_remove (UUID | dict[str, None | str | float | int]): UUID of the wellhead to remove or a dictionary \
            with sufficient matching parameters to uniquely identify a wellhead

        """
        self.__remove_object_operations.remove_object_from_network_main(
            obj_to_remove, self._network_element_name, self.__wellheads)

    def add(self, obj_to_remove: dict[str, None | str | float | int]) -> None:
        """Adds a wellhead to a network, taking a dictionary with properties for the new wellhead.

        Args:
            obj_to_remove (dict[str, None | str | float | int]): dictionary taking all the properties for the new
            wellhead.
            Requires date and a name.
        """
        new_object = self.__add_object_operations.add_network_obj(obj_to_remove, NexusWellhead, self.__parent_network)
        self._add_to_memory([new_object])

    def modify(self, obj_to_modify: dict[str, None | str | float | int],
               new_properties: dict[str, None | str | float | int]) -> None:
        """Modifies an existing wellhead based on a matching dictionary of properties.
        (partial matches allowed if precisely 1 matching node is found).
        Updates the properties with properties in the new_properties dictionary.

        Args:
            obj_to_modify (dict[str, None | str | float | int]): dictionary containing attributes to match in the
            existing wellheads.
            new_properties (dict[str, None | str | float | int]): properties to switch to in the new wellhead
        """
        self.__parent_network.get_load_status()
        self.__modify_object_operations.modify_network_object(obj_to_modify, new_properties,
                                                              self.__parent_network)

    @property
    def table_header(self) -> str:
        """Start of the wellhead definition table."""
        return 'WELLHEAD'

    @property
    def table_footer(self) -> str:
        """End of the wellhead definition table."""
        return 'END' + self.table_header



"""Unit types for the attributes of a well connection, wellhead, wellbore, etc."""
from typing import Mapping

from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.Units.AttributeMappings.BaseUnitMapping import BaseUnitMapping
from ResSimpy.Units.Units import (UnitDimension, Length, Temperature, Diameter, Roughness,
                                  Dimensionless, HeatTransfer, Time, DeltaPressure, ProductivityIndex,
                                  NonDarcySkin, ThermalConductivity)


class NetworkUnits(BaseUnitMapping):
    """Unit types for the attributes of a well connection, wellhead, wellbore, etc."""

    def __init__(self, unit_system: None | UnitSystem) -> None:
        super().__init__(unit_system=unit_system)
        self.attribute_map: Mapping[str, UnitDimension] = {
            'bhdepth': Length(),
            'depth': Length(),
            'datum_depth': Length(),
            'x_pos': Length(),
            'y_pos': Length(),
            'length': Length(),
            'temperature': Temperature(),
            'diameter': Diameter(),
            'roughness': Roughness(),
            'inner_diameter': Diameter(),
            'productivity_index': ProductivityIndex(),
            'rate_mult': Dimensionless(),
            'dp_add': DeltaPressure(),
            'dt_add': Temperature(),
            'well_index_mult': Dimensionless(),
            'vip_productivity_index': ProductivityIndex(),
            'd_factor': NonDarcySkin(),
            'gas_mobility': Dimensionless(),
            'drill_order_benefit': Dimensionless(),
            'heat_transfer_coeff': HeatTransfer(),
            'non_darcy_flow_model': Dimensionless(),
            'non_darcy_flow_method': Dimensionless(),
            'capillary_number_model': Dimensionless(),
            'on_time': Dimensionless(),
            'crossflow': Dimensionless(),
            'crossshut': Dimensionless(),
            'inj_mobility': Dimensionless(),
            'temperature_profile': Dimensionless(),
            'elevation_profile': Dimensionless(),
            'measured_depth_in': Length(),
            'measured_depth_out': Length(),
            'delta_depth': Length(),
            'temp': Temperature(),
            'hyd_method': Dimensionless(),
            'pvt_method': Dimensionless(),
            'water_method': Dimensionless(),
            'bat_method': Dimensionless(),
            'number': Dimensionless(),
            'gradient_calc': Dimensionless(),
            'add_tubing': Dimensionless(),
            'tracer': Dimensionless(),
            'con_type': Dimensionless(),
            'bore_type': Dimensionless(),

            # Units from Target class
            'control_quantity': Dimensionless(),
            'control_type': Dimensionless(),
            'control_condition': Dimensionless(),
            'control_method': Dimensionless(),
            'control_connections': Dimensionless(),
            'calculation_method': Dimensionless(),
            'calculation_conditions': Dimensionless(),
            'calculation_connections': Dimensionless(),
            'value': Dimensionless(),
            'add_value': Dimensionless(),
            'region': Dimensionless(),
            'priority': Dimensionless(),
            'minimum_rate': Dimensionless(),
            'minimum_rate_no_shut': Dimensionless(),
            'guide_rate': Dimensionless(),
            'max_change_pressure': DeltaPressure(),
            'rank_dt': Time(),
            'calculation_type': Dimensionless(),
            'temperature_in': Temperature(),
            'temperature_out': Temperature(),
            'heat_conductivity': ThermalConductivity(),
            'insulation_thickness': Diameter(),  # Same units as diameter but is annular
            'insulation_conductivity': ThermalConductivity(),
            'gravity_pressure_gradient_mult': Dimensionless(),
            'friction_pressure_gradient_mult': Dimensionless(),
            'acceleration_pressure_gradient_mult': Dimensionless(),
        }

    @property
    def bhdepth(self) -> str:
        """Returns the unit for bhdepth."""
        return self.get_unit_for_attribute('bhdepth')

    @property
    def depth(self) -> str:
        """Returns the unit for depth."""
        return self.get_unit_for_attribute('depth')

    @property
    def datum_depth(self) -> str:
        """Returns the unit for datum_depth."""
        return self.get_unit_for_attribute('datum_depth')

    @property
    def x_pos(self) -> str:
        """Returns the unit for x_pos."""
        return self.get_unit_for_attribute('x_pos')

    @property
    def y_pos(self) -> str:
        """Returns the unit for y_pos."""
        return self.get_unit_for_attribute('y_pos')

    @property
    def length(self) -> str:
        """Returns the unit for length."""
        return self.get_unit_for_attribute('length')

    @property
    def temperature(self) -> str:
        """Returns the unit for temperature."""
        return self.get_unit_for_attribute('temperature')

    @property
    def diameter(self) -> str:
        """Returns the unit for diameter."""
        return self.get_unit_for_attribute('diameter')

    @property
    def roughness(self) -> str:
        """Returns the unit for roughness."""
        return self.get_unit_for_attribute('roughness')

    @property
    def inner_diameter(self) -> str:
        """Returns the unit for inner_diameter."""
        return self.get_unit_for_attribute('inner_diameter')

    @property
    def productivity_index(self) -> str:
        """Returns the unit for productivity_index."""
        return self.get_unit_for_attribute('productivity_index')

    @property
    def rate_mult(self) -> str:
        """Returns the unit for rate_mult."""
        return self.get_unit_for_attribute('rate_mult')

    @property
    def dp_add(self) -> str:
        """Returns the unit for dp_add."""
        return self.get_unit_for_attribute('dp_add')

    @property
    def dt_add(self) -> str:
        """Returns the unit for dt_add."""
        return self.get_unit_for_attribute('dt_add')

    @property
    def well_index_mult(self) -> str:
        """Returns the unit for well_index_mult."""
        return self.get_unit_for_attribute('well_index_mult')

    @property
    def vip_productivity_index(self) -> str:
        """Returns the unit for vip_productivity_index."""
        return self.get_unit_for_attribute('vip_productivity_index')

    @property
    def d_factor(self) -> str:
        """Returns the unit for d_factor."""
        return self.get_unit_for_attribute('d_factor')

    @property
    def gas_mobility(self) -> str:
        """Returns the unit for gas_mobility."""
        return self.get_unit_for_attribute('gas_mobility')

    @property
    def drill_order_benefit(self) -> str:
        """Returns the unit for drill_order_benefit."""
        return self.get_unit_for_attribute('drill_order_benefit')

    @property
    def heat_transfer_coeff(self) -> str:
        """Returns the unit for heat_transfer_coeff."""
        return self.get_unit_for_attribute('heat_transfer_coeff')

    @property
    def non_darcy_flow_model(self) -> str:
        """Returns the unit for non_darcy_flow_model."""
        return self.get_unit_for_attribute('non_darcy_flow_model')

    @property
    def non_darcy_flow_method(self) -> str:
        """Returns the unit for non_darcy_flow_method."""
        return self.get_unit_for_attribute('non_darcy_flow_method')

    @property
    def capillary_number_model(self) -> str:
        """Returns the unit for capillary_number_model."""
        return self.get_unit_for_attribute('capillary_number_model')

    @property
    def on_time(self) -> str:
        """Returns the unit for on_time."""
        return self.get_unit_for_attribute('on_time')

    @property
    def crossflow(self) -> str:
        """Returns the unit for crossflow."""
        return self.get_unit_for_attribute('crossflow')

    @property
    def crossshut(self) -> str:
        """Returns the unit for crossshut."""
        return self.get_unit_for_attribute('crossshut')

    @property
    def inj_mobility(self) -> str:
        """Returns the unit for inj_mobility."""
        return self.get_unit_for_attribute('inj_mobility')

    @property
    def temperature_profile(self) -> str:
        """Returns the unit for temperature_profile."""
        return self.get_unit_for_attribute('temperature_profile')

    @property
    def elevation_profile(self) -> str:
        """Returns the unit for elevation_profile."""
        return self.get_unit_for_attribute('elevation_profile')

    @property
    def measured_depth_in(self) -> str:
        """Returns the unit for measured_depth_in."""
        return self.get_unit_for_attribute('measured_depth_in')

    @property
    def measured_depth_out(self) -> str:
        """Returns the unit for measured_depth_out."""
        return self.get_unit_for_attribute('measured_depth_out')

    @property
    def delta_depth(self) -> str:
        """Returns the unit for delta_depth."""
        return self.get_unit_for_attribute('delta_depth')

    @property
    def temp(self) -> str:
        """Returns the unit for temp."""
        return self.get_unit_for_attribute('temp')

    @property
    def hyd_method(self) -> str:
        """Returns the unit for hyd_method."""
        return self.get_unit_for_attribute('hyd_method')

    @property
    def pvt_method(self) -> str:
        """Returns the unit for pvt_method."""
        return self.get_unit_for_attribute('pvt_method')

    @property
    def water_method(self) -> str:
        """Returns the unit for water_method."""
        return self.get_unit_for_attribute('water_method')

    @property
    def bat_method(self) -> str:
        """Returns the unit for bat_method."""
        return self.get_unit_for_attribute('bat_method')

    @property
    def number(self) -> str:
        """Returns the unit for number."""
        return self.get_unit_for_attribute('number')

    @property
    def gradient_calc(self) -> str:
        """Returns the unit for gradient_calc."""
        return self.get_unit_for_attribute('gradient_calc')

    @property
    def add_tubing(self) -> str:
        """Returns the unit for add_tubing."""
        return self.get_unit_for_attribute('add_tubing')

    @property
    def tracer(self) -> str:
        """Returns the unit for tracer."""
        return self.get_unit_for_attribute('tracer')

    @property
    def con_type(self) -> str:
        """Returns the unit for con_type."""
        return self.get_unit_for_attribute('con_type')

    @property
    def bore_type(self) -> str:
        """Returns the unit for bore_type."""
        return self.get_unit_for_attribute('bore_type')

    @property
    def control_quantity(self) -> str:
        """Returns the unit for control_quantity."""
        return self.get_unit_for_attribute('control_quantity')

    @property
    def control_type(self) -> str:
        """Returns the unit for control_type."""
        return self.get_unit_for_attribute('control_type')

    @property
    def control_condition(self) -> str:
        """Returns the unit for control_condition."""
        return self.get_unit_for_attribute('control_condition')

    @property
    def control_method(self) -> str:
        """Returns the unit for control_method."""
        return self.get_unit_for_attribute('control_method')

    @property
    def control_connections(self) -> str:
        """Returns the unit for control_connections."""
        return self.get_unit_for_attribute('control_connections')

    @property
    def calculation_method(self) -> str:
        """Returns the unit for calculation_method."""
        return self.get_unit_for_attribute('calculation_method')

    @property
    def calculation_conditions(self) -> str:
        """Returns the unit for calculation_conditions."""
        return self.get_unit_for_attribute('calculation_conditions')

    @property
    def calculation_connections(self) -> str:
        """Returns the unit for calculation_connections."""
        return self.get_unit_for_attribute('calculation_connections')

    @property
    def value(self) -> str:
        """Returns the unit for value."""
        return self.get_unit_for_attribute('value')

    @property
    def add_value(self) -> str:
        """Returns the unit for add_value."""
        return self.get_unit_for_attribute('add_value')

    @property
    def region(self) -> str:
        """Returns the unit for region."""
        return self.get_unit_for_attribute('region')

    @property
    def priority(self) -> str:
        """Returns the unit for priority."""
        return self.get_unit_for_attribute('priority')

    @property
    def minimum_rate(self) -> str:
        """Returns the unit for minimum_rate."""
        return self.get_unit_for_attribute('minimum_rate')

    @property
    def minimum_rate_no_shut(self) -> str:
        """Returns the unit for minimum_rate_no_shut."""
        return self.get_unit_for_attribute('minimum_rate_no_shut')

    @property
    def guide_rate(self) -> str:
        """Returns the unit for guide_rate."""
        return self.get_unit_for_attribute('guide_rate')

    @property
    def max_change_pressure(self) -> str:
        """Returns the unit for max_change_pressure."""
        return self.get_unit_for_attribute('max_change_pressure')

    @property
    def rank_dt(self) -> str:
        """Returns the unit for rank_dt."""
        return self.get_unit_for_attribute('rank_dt')

    @property
    def calculation_type(self) -> str:
        """Returns the unit for calculation_type."""
        return self.get_unit_for_attribute('calculation_type')

    @property
    def temperature_in(self) -> str:
        """Returns the unit for temperature_in."""
        return self.get_unit_for_attribute('temperature_in')

    @property
    def temperature_out(self) -> str:
        """Returns the unit for temperature_out."""
        return self.get_unit_for_attribute('temperature_out')

    @property
    def heat_conductivity(self) -> str:
        """Returns the unit for heat_conductivity."""
        return self.get_unit_for_attribute('heat_conductivity')

    @property
    def insulation_thickness(self) -> str:
        """Returns the unit for insulation_thickness."""
        return self.get_unit_for_attribute('insulation_thickness')

    @property
    def insulation_conductivity(self) -> str:
        """Returns the unit for insulation_conductivity."""
        return self.get_unit_for_attribute('insulation_conductivity')

    @property
    def gravity_pressure_gradient_mult(self) -> str:
        """Returns the unit for gravity_pressure_gradient_mult."""
        return self.get_unit_for_attribute('gravity_pressure_gradient_mult')

    @property
    def friction_pressure_gradient_mult(self) -> str:
        """Returns the unit for friction_pressure_gradient_mult."""
        return self.get_unit_for_attribute('friction_pressure_gradient_mult')

    @property
    def acceleration_pressure_gradient_mult(self) -> str:
        """Returns the unit for acceleration_pressure_gradient_mult."""
        return self.get_unit_for_attribute('acceleration_pressure_gradient_mult')



from typing import Mapping
from ResSimpy.Enums.UnitsEnum import UnitSystem

from ResSimpy.Units.AttributeMappings.BaseUnitMapping import BaseUnitMapping
from ResSimpy.Units.Units import (HeatCapacity, UnitDimension, Viscosity, Density, GasLiquidRatio, LiquidGasRatio,
                                  SolutionOilGasRatio, DeltaPressure, Temperature, ReservoirVolumeOverPressure,
                                  FormationVolumeFactorGas, FormationVolumeFactorLiquid, SolutionGasOilRatio,
                                  Dimensionless, Pressure, CriticalPressure, CriticalTemperature, CriticalVolume,
                                  Compressibility, Length, InverseTime, Permeability, ReservoirProductivityIndex,
                                  ReservoirVolume, SurfaceRatesLiquid, SurfaceRatesGas, GravityGradient,
                                  ValveCoefficient, Roughness, Diameter, InterfacialTension
                                  )


class WaterUnits(BaseUnitMapping):
    """Unit types for the attributes of water methods."""

    def __init__(self, unit_system: None | UnitSystem) -> None:
        super().__init__(unit_system=unit_system)

    attribute_map: Mapping[str, UnitDimension] = {
        'temperature': Temperature(),
        'reference_pressure': Pressure(),
        'water_density': Density(),
        'water_compressibility': Compressibility(),
        'water_formation_volume_factor': FormationVolumeFactorLiquid(),
        'water_viscosity': Viscosity(),
        'water_viscosity_compressibility': Compressibility()
    }

    @property
    def temperature(self) -> str:
        """Returns the unit for temperature."""
        return self.get_unit_for_attribute('temperature')

    @property
    def reference_pressure(self) -> str:
        """Returns the unit for reference_pressure."""
        return self.get_unit_for_attribute('reference_pressure')

    @property
    def water_density(self) -> str:
        """Returns the unit for water_density."""
        return self.get_unit_for_attribute('water_density')

    @property
    def water_compressibility(self) -> str:
        """Returns the unit for water_compressibility."""
        return self.get_unit_for_attribute('water_compressibility')

    @property
    def water_formation_volume_factor(self) -> str:
        """Returns the unit for water_formation_volume_factor."""
        return self.get_unit_for_attribute('water_formation_volume_factor')

    @property
    def water_viscosity(self) -> str:
        """Returns the unit for water_viscosity."""
        return self.get_unit_for_attribute('water_viscosity')

    @property
    def water_viscosity_compressibility(self) -> str:
        """Returns the unit for water_viscosity_compressibility."""
        return self.get_unit_for_attribute('water_viscosity_compressibility')


class SeparatorUnits(BaseUnitMapping):
    """Unit types for the attributes of separator methods."""

    def __init__(self, unit_system: None | UnitSystem) -> None:
        super().__init__(unit_system=unit_system)

    attribute_map: Mapping[str, UnitDimension] = {
        'temperature': Temperature(),
        'pressure': Pressure(),
        'standard_temperature': Temperature(),
        'standard_pressure': Pressure(),
    }

    @property
    def temperature(self) -> str:
        """Returns the unit for temperature."""
        return self.get_unit_for_attribute('temperature')

    @property
    def pressure(self) -> str:
        """Returns the unit for pressure."""
        return self.get_unit_for_attribute('pressure')

    @property
    def standard_temperature(self) -> str:
        """Returns the unit for standard_temperature."""
        return self.get_unit_for_attribute('standard_temperature')

    @property
    def standard_pressure(self) -> str:
        """Returns the unit for standard_pressure."""
        return self.get_unit_for_attribute('standard_pressure')


class RockUnits(BaseUnitMapping):
    """Unit types for the attributes of rock methods."""

    def __init__(self, unit_system: None | UnitSystem) -> None:
        super().__init__(unit_system=unit_system)

    attribute_map: Mapping[str, UnitDimension] = {
        'rock_compressibility': Compressibility(),
        'rock_permeability_compressibility': Compressibility(),
        'reference_pressure': Pressure(),
        'pressure': Pressure(),
        'delta_pressure': DeltaPressure()
    }

    @property
    def rock_compressibility(self) -> str:
        """Returns the unit for rock_compressibility."""
        return self.get_unit_for_attribute('rock_compressibility')

    @property
    def rock_permeability_compressibility(self) -> str:
        """Returns the unit for rock_permeability_compressibility."""
        return self.get_unit_for_attribute('rock_permeability_compressibility')

    @property
    def reference_pressure(self) -> str:
        """Returns the unit for reference_pressure."""
        return self.get_unit_for_attribute('reference_pressure')

    @property
    def pressure(self) -> str:
        """Returns the unit for pressure."""
        return self.get_unit_for_attribute('pressure')

    @property
    def delta_pressure(self) -> str:
        """Returns the unit for delta_pressure."""
        return self.get_unit_for_attribute('delta_pressure')


class RelPermUnits(BaseUnitMapping):
    """Unit types for the attributes of relative permeability and capillary pressure methods."""

    def __init__(self, unit_system: None | UnitSystem) -> None:
        super().__init__(unit_system=unit_system)

    attribute_map: Mapping[str, UnitDimension] = {
        'water_oil_capillary_pressure': Pressure(),
        'gas_oil_capillary_pressure': Pressure(),
        'gas_water_capillary_pressure': Pressure(),
        'interfacial_tension_threshold_for_relperm_adjustment': InterfacialTension(),
        'reference_interfacial_tension_for_capillary_pressure_adjustment': InterfacialTension()
    }

    @property
    def water_oil_capillary_pressure(self) -> str:
        """Returns the unit for water_oil_capillary_pressure."""
        return self.get_unit_for_attribute('water_oil_capillary_pressure')

    @property
    def gas_oil_capillary_pressure(self) -> str:
        """Returns the unit for gas_oil_capillary_pressure."""
        return self.get_unit_for_attribute('gas_oil_capillary_pressure')

    @property
    def gas_water_capillary_pressure(self) -> str:
        """Returns the unit for gas_water_capillary_pressure."""
        return self.get_unit_for_attribute('gas_water_capillary_pressure')

    @property
    def interfacial_tension_threshold_for_relperm_adjustment(self) -> str:
        """Returns the unit for interfacial_tension_threshold_for_relperm_adjustment."""
        return self.get_unit_for_attribute('interfacial_tension_threshold_for_relperm_adjustment')

    @property
    def reference_interfacial_tension_for_capillary_pressure_adjustment(self) -> str:
        """Returns the unit for reference_interfacial_tension_for_capillary_pressure_adjustment."""
        return self.get_unit_for_attribute('reference_interfacial_tension_for_capillary_pressure_adjustment')


class HydraulicsUnits(BaseUnitMapping):
    """Unit types for the attributes of hydraulics, gaslift, valve, etc. methods."""

    def __init__(self, unit_system: None | UnitSystem, ratio_thousands: bool = True) -> None:
        self.ratio_thousands = ratio_thousands
        super().__init__(unit_system=unit_system)

    attribute_map: Mapping[str, UnitDimension] = {
        'surface_oil_rate': SurfaceRatesLiquid(),  # For use in hyd & gaslift methods
        'surface_liquid_rate': SurfaceRatesLiquid(),
        'surface_water_rate': SurfaceRatesLiquid(),
        'surface_gas_rate': SurfaceRatesGas(),
        'surface_wet_gas_rate': SurfaceRatesGas(),
        'mean_molecular_weight': Dimensionless(),
        'watercut': Dimensionless(),
        'oilcut': Dimensionless(),
        'pressure': Pressure(),
        'inlet_pressure': Pressure(),
        'outlet_pressure': Pressure(),
        'bottomhole_pressure': Pressure(),
        'tubinghead_pressure': Pressure(),
        'gas_liquid_ratio': GasLiquidRatio(),  # Note extra complexity in Nexus for VIP compatibility
        'gas_oil_ratio': SolutionGasOilRatio(),  # Note extra complexity in Nexus for VIP compatibility
        'oil_gas_ratio': LiquidGasRatio(),  # Note extra complexity in Nexus for VIP compatibility
        'water_gas_ratio': LiquidGasRatio(),  # Note extra complexity in Nexus for VIP compatibility
        'gas_water_ratio': GasLiquidRatio(),  # Note extra complexity in Nexus for VIP compatibility
        'water_wet_gas_ratio': LiquidGasRatio(),  # Note extra complexity in Nexus for VIP compatibility
        'length': Length(),
        'datum_depth': Length(),
        'depth_change': Length(),
        'hydraulic_table_vertical_distance': Length(),
        'injected_fluid_pressure_gradient': GravityGradient(),
        'viscosity': Viscosity(),
        'diameter': Diameter(),
        'roughness': Roughness(),
        'valve_coefficient': ValveCoefficient()  # For use in Valve method
    }

    @property
    def surface_oil_rate(self) -> str:
        """Returns the unit for surface_oil_rate."""
        return self.get_unit_for_attribute('surface_oil_rate')

    @property
    def surface_liquid_rate(self) -> str:
        """Returns the unit for surface_liquid_rate."""
        return self.get_unit_for_attribute('surface_liquid_rate')

    @property
    def surface_water_rate(self) -> str:
        """Returns the unit for surface_water_rate."""
        return self.get_unit_for_attribute('surface_water_rate')

    @property
    def surface_gas_rate(self) -> str:
        """Returns the unit for surface_gas_rate."""
        return self.get_unit_for_attribute('surface_gas_rate')

    @property
    def surface_wet_gas_rate(self) -> str:
        """Returns the unit for surface_wet_gas_rate."""
        return self.get_unit_for_attribute('surface_wet_gas_rate')

    @property
    def mean_molecular_weight(self) -> str:
        """Returns the unit for mean_molecular_weight."""
        return self.get_unit_for_attribute('mean_molecular_weight')

    @property
    def watercut(self) -> str:
        """Returns the unit for watercut."""
        return self.get_unit_for_attribute('watercut')

    @property
    def oilcut(self) -> str:
        """Returns the unit for oilcut."""
        return self.get_unit_for_attribute('oilcut')

    @property
    def pressure(self) -> str:
        """Returns the unit for pressure."""
        return self.get_unit_for_attribute('pressure')

    @property
    def inlet_pressure(self) -> str:
        """Returns the unit for inlet_pressure."""
        return self.get_unit_for_attribute('inlet_pressure')

    @property
    def outlet_pressure(self) -> str:
        """Returns the unit for outlet_pressure."""
        return self.get_unit_for_attribute('outlet_pressure')

    @property
    def bottomhole_pressure(self) -> str:
        """Returns the unit for bottomhole_pressure."""
        return self.get_unit_for_attribute('bottomhole_pressure')

    @property
    def tubinghead_pressure(self) -> str:
        """Returns the unit for tubinghead_pressure."""
        return self.get_unit_for_attribute('tubinghead_pressure')

    @property
    def gas_liquid_ratio(self) -> str:
        """Returns the unit for gas_liquid_ratio."""
        unit = self.get_unit_for_attribute('gas_liquid_ratio')
        if not self.ratio_thousands:
            unit = unit.replace('mscf', 'scf')
        return unit

    @property
    def gas_oil_ratio(self) -> str:
        """Returns the unit for gas_oil_ratio."""
        unit = self.get_unit_for_attribute('gas_oil_ratio')
        if not self.ratio_thousands:
            unit = unit.replace('mscf', 'scf')
        return unit

    @property
    def oil_gas_ratio(self) -> str:
        """Returns the unit for oil_gas_ratio."""
        unit = self.get_unit_for_attribute('oil_gas_ratio')
        if not self.ratio_thousands:
            unit = unit.replace('mscf', 'mmscf')
        return unit

    @property
    def water_gas_ratio(self) -> str:
        """Returns the unit for water_gas_ratio."""
        unit = self.get_unit_for_attribute('water_gas_ratio')
        if not self.ratio_thousands:
            unit = unit.replace('mscf', 'mmscf')
        return unit

    @property
    def gas_water_ratio(self) -> str:
        """Returns the unit for gas_water_ratio."""
        unit = self.get_unit_for_attribute('gas_water_ratio')
        if not self.ratio_thousands:
            unit = unit.replace('mscf', 'scf')
        return unit

    @property
    def water_wet_gas_ratio(self) -> str:
        """Returns the unit for water_wet_gas_ratio."""
        unit = self.get_unit_for_attribute('water_wet_gas_ratio')
        if not self.ratio_thousands:
            unit = unit.replace('mscf', 'mmscf')
        return unit

    @property
    def length(self) -> str:
        """Returns the unit for length."""
        return self.get_unit_for_attribute('length')

    @property
    def datum_depth(self) -> str:
        """Returns the unit for datum_depth."""
        return self.get_unit_for_attribute('datum_depth')

    @property
    def depth_change(self) -> str:
        """Returns the unit for depth_change."""
        return self.get_unit_for_attribute('depth_change')

    @property
    def hydraulic_table_vertical_distance(self) -> str:
        """Returns the unit for hydraulic_table_vertical_distance."""
        return self.get_unit_for_attribute('hydraulic_table_vertical_distance')

    @property
    def injected_fluid_pressure_gradient(self) -> str:
        """Returns the unit for injected_fluid_pressure_gradient."""
        return self.get_unit_for_attribute('injected_fluid_pressure_gradient')

    @property
    def viscosity(self) -> str:
        """Returns the unit for viscosity."""
        return self.get_unit_for_attribute('viscosity')

    @property
    def diameter(self) -> str:
        """Returns the unit for diameter."""
        return self.get_unit_for_attribute('diameter')

    @property
    def roughness(self) -> str:
        """Returns the unit for roughness."""
        return self.get_unit_for_attribute('roughness')

    @property
    def valve_coefficient(self) -> str:
        """Returns the unit for valve_coefficient."""
        return self.get_unit_for_attribute('valve_coefficient')


class EquilUnits(BaseUnitMapping):
    """Unit types for the attributes of equilibration methods."""

    def __init__(self, unit_system: None | UnitSystem) -> None:
        super().__init__(unit_system=unit_system)

    attribute_map: Mapping[str, UnitDimension] = {
        'initial_pressure': Pressure(),
        'datum_depth': Length(),
        'depth': Length(),
        'x': Length(),
        'y': Length(),
        'temperature': Temperature(),
        'initial_temperature': Temperature(),
        'gas_oil_contact_depth': Length(),
        'water_oil_contact_depth': Length(),
        'gas_water_contact_depth': Length(),
        'gas_oil_capillary_pressure_at_gas_oil_contact': Pressure(),
        'water_oil_capillary_pressure_at_water_oil_contact': Pressure(),
        'gas_water_capillary_pressure_at_gas_water_contact': Pressure(),
        'saturation_pressure': Pressure(),
        'oil_api_gravity': Dimensionless()
    }

    @property
    def initial_pressure(self) -> str:
        """Returns the unit for initial_pressure."""
        return self.get_unit_for_attribute('initial_pressure')

    @property
    def datum_depth(self) -> str:
        """Returns the unit for datum_depth."""
        return self.get_unit_for_attribute('datum_depth')

    @property
    def depth(self) -> str:
        """Returns the unit for depth."""
        return self.get_unit_for_attribute('depth')

    @property
    def x(self) -> str:
        """Returns the unit for x-coordinate."""
        return self.get_unit_for_attribute('x')

    @property
    def y(self) -> str:
        """Returns the unit for y-coordinate."""
        return self.get_unit_for_attribute('y')

    @property
    def temperature(self) -> str:
        """Returns the unit for temperature."""
        return self.get_unit_for_attribute('temperature')

    @property
    def initial_temperature(self) -> str:
        """Returns the unit for initial_temperature."""
        return self.get_unit_for_attribute('initial_temperature')

    @property
    def gas_oil_contact_depth(self) -> str:
        """Returns the unit for gas_oil_contact_depth."""
        return self.get_unit_for_attribute('gas_oil_contact_depth')

    @property
    def water_oil_contact_depth(self) -> str:
        """Returns the unit for water_oil_contact_depth."""
        return self.get_unit_for_attribute('water_oil_contact_depth')

    @property
    def gas_water_contact_depth(self) -> str:
        """Returns the unit for gas_water_contact_depth."""
        return self.get_unit_for_attribute('gas_water_contact_depth')

    @property
    def gas_oil_capillary_pressure_at_gas_oil_contact(self) -> str:
        """Returns the unit for gas_oil_capillary_pressure_at_gas_oil_contact."""
        return self.get_unit_for_attribute('gas_oil_capillary_pressure_at_gas_oil_contact')

    @property
    def water_oil_capillary_pressure_at_water_oil_contact(self) -> str:
        """Returns the unit for water_oil_capillary_pressure_at_water_oil_contact."""
        return self.get_unit_for_attribute('water_oil_capillary_pressure_at_water_oil_contact')

    @property
    def gas_water_capillary_pressure_at_gas_water_contact(self) -> str:
        """Returns the unit for gas_water_capillary_pressure_at_gas_water_contact."""
        return self.get_unit_for_attribute('gas_water_capillary_pressure_at_gas_water_contact')

    @property
    def saturation_pressure(self) -> str:
        """Returns the unit for saturation_pressure."""
        return self.get_unit_for_attribute('saturation_pressure')

    @property
    def oil_api_gravity(self) -> str:
        """Returns the unit for oil_api_gravity."""
        return self.get_unit_for_attribute('oil_api_gravity')


class AquiferUnits(BaseUnitMapping):
    """Unit types for the attributes of Aquifer methods."""

    def __init__(self, unit_system: None | UnitSystem) -> None:
        super().__init__(unit_system=unit_system)

    attribute_map: Mapping[str, UnitDimension] = {
        'carter_tracy_constant': ReservoirVolumeOverPressure(),
        'total_compressibility': Compressibility(),
        'porosity': Dimensionless(),
        'thickness': Length(),
        'radius_to_inner_perimeter': Length(),
        'radius_to_exterior_perimeter': Length(),
        'fraction_of_circular_boundary': Dimensionless(),
        'linear_aquifer_width': Length(),
        'linear_aquifer_length': Length(),
        'time_conversion_factor': InverseTime(),
        'viscosity': Viscosity(),
        'permeability': Permeability(),
        'initial_aquifer_pressure': Pressure(),
        'datum_depth': Length(),
        'productivity_index': ReservoirProductivityIndex(),
        'initial_encroachable_water_volume': ReservoirVolume(),
        'initial_aquifer_volume': ReservoirVolume()
    }

    @property
    def carter_tracy_constant(self) -> str:
        """Returns the unit for carter_tracy_constant."""
        return self.get_unit_for_attribute('carter_tracy_constant')

    @property
    def total_compressibility(self) -> str:
        """Returns the unit for total_compressibility."""
        return self.get_unit_for_attribute('total_compressibility')

    @property
    def porosity(self) -> str:
        """Returns the unit for porosity."""
        return self.get_unit_for_attribute('porosity')

    @property
    def thickness(self) -> str:
        """Returns the unit for thickness."""
        return self.get_unit_for_attribute('thickness')

    @property
    def radius_to_inner_perimeter(self) -> str:
        """Returns the unit for radius_to_inner_perimeter."""
        return self.get_unit_for_attribute('radius_to_inner_perimeter')

    @property
    def radius_to_exterior_perimeter(self) -> str:
        """Returns the unit for radius_to_exterior_perimeter."""
        return self.get_unit_for_attribute('radius_to_exterior_perimeter')

    @property
    def fraction_of_circular_boundary(self) -> str:
        """Returns the unit for fraction_of_circular_boundary."""
        return self.get_unit_for_attribute('fraction_of_circular_boundary')

    @property
    def linear_aquifer_width(self) -> str:
        """Returns the unit for linear_aquifer_width."""
        return self.get_unit_for_attribute('linear_aquifer_width')

    @property
    def linear_aquifer_length(self) -> str:
        """Returns the unit for linear_aquifer_length."""
        return self.get_unit_for_attribute('linear_aquifer_length')

    @property
    def time_conversion_factor(self) -> str:
        """Returns the unit for time_conversion_factor."""
        return self.get_unit_for_attribute('time_conversion_factor')

    @property
    def viscosity(self) -> str:
        """Returns the unit for viscosity."""
        return self.get_unit_for_attribute('viscosity')

    @property
    def permeability(self) -> str:
        """Returns the unit for permeability."""
        return self.get_unit_for_attribute('permeability')

    @property
    def initial_aquifer_pressure(self) -> str:
        """Returns the unit for initial_aquifer_pressure."""
        return self.get_unit_for_attribute('initial_aquifer_pressure')

    @property
    def datum_depth(self) -> str:
        """Returns the unit for datum_depth."""
        return self.get_unit_for_attribute('datum_depth')

    @property
    def productivity_index(self) -> str:
        """Returns the unit for productivity_index."""
        return self.get_unit_for_attribute('productivity_index')

    @property
    def initial_encroachable_water_volume(self) -> str:
        """Returns the unit for initial_encroachable_water_volume."""
        return self.get_unit_for_attribute('initial_encroachable_water_volume')

    @property
    def initial_aquifer_volume(self) -> str:
        """Returns the unit for initial_aquifer_volume."""
        return self.get_unit_for_attribute('initial_aquifer_volume')


class PVTUnits(BaseUnitMapping):
    """Unit types for the attributes of PVT methods."""

    def __init__(self, unit_system: None | UnitSystem) -> None:
        super().__init__(unit_system=unit_system)

    attribute_map: Mapping[str, UnitDimension] = {
        'pressure': Pressure(),
        'temperature': Temperature(),
        'delta_pressure': DeltaPressure(),
        'oil_density': Density(),
        'gas_density': Density(),
        'oil_api_gravity': Dimensionless(),
        'gas_specific_gravity': Dimensionless(),
        'molecular_weight_of_residual_oil': Dimensionless(),
        'oil_viscosity': Viscosity(),
        'gas_viscosity': Viscosity(),
        'oil_formation_volume_factor': FormationVolumeFactorLiquid(),
        'gas_formation_volume_factor': FormationVolumeFactorGas(),
        'solution_gas_oil_ratio': SolutionGasOilRatio(),
        'solution_oil_gas_ratio': SolutionOilGasRatio(),
        'saturation_pressure': Pressure(),
        'gas_oil_ratio': GasLiquidRatio(),
        'oil_gas_ratio': LiquidGasRatio(),
        'gas_heat_capacity_at_constant_volume': HeatCapacity(),
        'gas_heat_capacity_at_constant_pressure': HeatCapacity(),
        'oil_fvf_over_oil_fvf_at_saturation_pressure': Dimensionless(),
        'oil_viscosity_over_oil_viscosity_at_saturation_pressure': Dimensionless(),
        'gas_fvf_over_gas_fvf_at_saturation_pressure': Dimensionless(),
        'gas_viscosity_over_gas_viscosity_at_saturation_pressure': Dimensionless(),
        'gas_heat_capacity_at_constant_volume_over_cv_at_saturation_pressure': Dimensionless(),
        'gas_heat_capacity_at_constant_pressure_over_cp_at_saturation_pressure': Dimensionless(),
        'critical_pressure': CriticalPressure(),
        'critical_temperature': CriticalTemperature(),
        'critical_volume': CriticalVolume()
    }

    @property
    def pressure(self) -> str:
        """Returns the unit for pressure."""
        return self.get_unit_for_attribute('pressure')

    @property
    def temperature(self) -> str:
        """Returns the unit for temperature."""
        return self.get_unit_for_attribute('temperature')

    @property
    def delta_pressure(self) -> str:
        """Returns the unit for delta_pressure."""
        return self.get_unit_for_attribute('delta_pressure')

    @property
    def oil_density(self) -> str:
        """Returns the unit for oil_density."""
        return self.get_unit_for_attribute('oil_density')

    @property
    def gas_density(self) -> str:
        """Returns the unit for gas_density."""
        return self.get_unit_for_attribute('gas_density')

    @property
    def oil_api_gravity(self) -> str:
        """Returns the unit for oil_api_gravity."""
        return self.get_unit_for_attribute('oil_api_gravity')

    @property
    def gas_specific_gravity(self) -> str:
        """Returns the unit for gas_specific_gravity."""
        return self.get_unit_for_attribute('gas_specific_gravity')

    @property
    def molecular_weight_of_residual_oil(self) -> str:
        """Returns the unit for molecular_weight_of_residual_oil."""
        return self.get_unit_for_attribute('molecular_weight_of_residual_oil')

    @property
    def oil_viscosity(self) -> str:
        """Returns the unit for oil_viscosity."""
        return self.get_unit_for_attribute('oil_viscosity')

    @property
    def gas_viscosity(self) -> str:
        """Returns the unit for gas_viscosity."""
        return self.get_unit_for_attribute('gas_viscosity')

    @property
    def oil_formation_volume_factor(self) -> str:
        """Returns the unit for oil_formation_volume_factor."""
        return self.get_unit_for_attribute('oil_formation_volume_factor')

    @property
    def gas_formation_volume_factor(self) -> str:
        """Returns the unit for gas_formation_volume_factor."""
        return self.get_unit_for_attribute('gas_formation_volume_factor')

    @property
    def solution_gas_oil_ratio(self) -> str:
        """Returns the unit for solution_gas_oil_ratio."""
        return self.get_unit_for_attribute('solution_gas_oil_ratio')

    @property
    def solution_oil_gas_ratio(self) -> str:
        """Returns the unit for solution_oil_gas_ratio."""
        return self.get_unit_for_attribute('solution_oil_gas_ratio')

    @property
    def saturation_pressure(self) -> str:
        """Returns the unit for saturation_pressure."""
        return self.get_unit_for_attribute('saturation_pressure')

    @property
    def gas_oil_ratio(self) -> str:
        """Returns the unit for gas_oil_ratio."""
        return self.get_unit_for_attribute('gas_oil_ratio')

    @property
    def oil_gas_ratio(self) -> str:
        """Returns the unit for oil_gas_ratio."""
        return self.get_unit_for_attribute('oil_gas_ratio')

    @property
    def gas_heat_capacity_at_constant_volume(self) -> str:
        """Returns the unit for gas_heat_capacity_at_constant_volume."""
        return self.get_unit_for_attribute('gas_heat_capacity_at_constant_volume')

    @property
    def gas_heat_capacity_at_constant_pressure(self) -> str:
        """Returns the unit for gas_heat_capacity_at_constant_pressure."""
        return self.get_unit_for_attribute('gas_heat_capacity_at_constant_pressure')

    @property
    def oil_fvf_over_oil_fvf_at_saturation_pressure(self) -> str:
        """Returns the unit for oil_fvf_over_oil_fvf_at_saturation_pressure."""
        return self.get_unit_for_attribute('oil_fvf_over_oil_fvf_at_saturation_pressure')

    @property
    def oil_viscosity_over_oil_viscosity_at_saturation_pressure(self) -> str:
        """Returns the unit for oil_viscosity_over_oil_viscosity_at_saturation_pressure."""
        return self.get_unit_for_attribute('oil_viscosity_over_oil_viscosity_at_saturation_pressure')

    @property
    def gas_fvf_over_gas_fvf_at_saturation_pressure(self) -> str:
        """Returns the unit for gas_fvf_over_gas_fvf_at_saturation_pressure."""
        return self.get_unit_for_attribute('gas_fvf_over_gas_fvf_at_saturation_pressure')

    @property
    def gas_viscosity_over_gas_viscosity_at_saturation_pressure(self) -> str:
        """Returns the unit for gas_viscosity_over_gas_viscosity_at_saturation_pressure."""
        return self.get_unit_for_attribute('gas_viscosity_over_gas_viscosity_at_saturation_pressure')

    @property
    def gas_heat_capacity_at_constant_volume_over_cv_at_saturation_pressure(self) -> str:
        """Returns the unit for gas_heat_capacity_at_constant_volume_over_cv_at_saturation_pressure."""
        return self.get_unit_for_attribute('gas_heat_capacity_at_constant_volume_over_cv_at_saturation_pressure')

    @property
    def gas_heat_capacity_at_constant_pressure_over_cp_at_saturation_pressure(self) -> str:
        """Returns the unit for gas_heat_capacity_at_constant_pressure_over_cp_at_saturation_pressure."""
        return self.get_unit_for_attribute('gas_heat_capacity_at_constant_pressure_over_cp_at_saturation_pressure')

    @property
    def critical_pressure(self) -> str:
        """Returns the unit for critical_pressure."""
        return self.get_unit_for_attribute('critical_pressure')

    @property
    def critical_temperature(self) -> str:
        """Returns the unit for critical_temperature."""
        return self.get_unit_for_attribute('critical_temperature')

    @property
    def critical_volume(self) -> str:
        """Returns the unit for critical_volume."""
        return self.get_unit_for_attribute('critical_volume')



"""Abstract base class for constraints."""
from __future__ import annotations
from dataclasses import dataclass, field
from abc import ABC, abstractmethod
import pandas as pd

from ResSimpy.Constraint import Constraint
from typing import Optional, Mapping, Sequence

from ResSimpy.Enums.UnitsEnum import UnitSystem


@dataclass(kw_only=True)
class Constraints(ABC):
    """Abstract base class for constraints."""
    __constraints: dict[str, list[Constraint]] = field(default_factory=lambda: {})

    @abstractmethod
    def get_all(self, object_name: Optional[str] = None, date: Optional[str] = None) \
            -> Mapping[str, Sequence[Constraint]]:
        raise NotImplementedError("Implement this in the derived class")

    @abstractmethod
    def get_df(self) -> pd.DataFrame:
        raise NotImplementedError("Implement this in the derived class")

    @abstractmethod
    def get_overview(self) -> str:
        raise NotImplementedError("Implement this in the derived class")

    @abstractmethod
    def remove(self) -> None:
        raise NotImplementedError("Implement this in the derived class")

    @abstractmethod
    def add(self,
            constraint_to_add: dict[str, None | float | int | str | UnitSystem] | Constraint,
            comments: str | None,
            ) -> None:
        raise NotImplementedError("Implement this in the derived class")

    @abstractmethod
    def modify(self, name: str,
               current_constraint: dict[str, None | float | int | str] | Constraint,
               new_constraint_props: dict[str, None | float | int | str | UnitSystem] | Constraint,
               comments: Optional[str] = None) \
            -> None:
        raise NotImplementedError("Implement this in the derived class")



"""Base class object for storing data related to well and node constraints."""

from __future__ import annotations
from abc import ABC
from dataclasses import dataclass
from typing import Optional

from ResSimpy.DataObjectMixin import DataObjectMixin
from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.Units.AttributeMappings.ConstraintUnitMapping import ConstraintUnits


@dataclass(repr=False)
class Constraint(DataObjectMixin, ABC):
    """Base class object for storing data related to well and node constraints."""
    name: Optional[str] = None
    date: Optional[str] = None
    unit_system: Optional[UnitSystem] = None
    max_surface_oil_rate: Optional[float] = None
    max_surface_gas_rate: Optional[float] = None
    max_surface_water_rate: Optional[float] = None
    max_surface_liquid_rate: Optional[float] = None
    max_reservoir_oil_rate: Optional[float] = None
    max_reservoir_gas_rate: Optional[float] = None
    max_reservoir_water_rate: Optional[float] = None
    max_reservoir_liquid_rate: Optional[float] = None

    @property
    def units(self) -> ConstraintUnits:
        """Returns the attribute to unit map for the constraint."""
        return ConstraintUnits(self.unit_system)
